# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re

PATH_WHITE_LIST_REGEX = re.compile(r"[^_A-Za-z0-9/.-]")

CONFIG_FILE_MAX_SIZE = 1 * 1024 * 1024  # work for .ini config file
TEXT_FILE_MAX_SIZE = 100 * 1024 * 1024  # work for txt, py
CSV_FILE_MAX_SIZE = 1024 * 1024 * 1024
JSON_FILE_MAX_SIZE = 1024 * 1024 * 1024
ONNX_MODEL_MAX_SIZE = 2 * 1024 * 1024 * 1024
TENSOR_MAX_SIZE = 10 * 1024 * 1024 * 1024
MODEL_WEIGHT_MAX_SIZE = 300 * 1024 * 1024 * 1024
INPUT_FILE_MAX_SIZE = 5 * 1024 * 1024 * 1024
LOG_FILE_MAX_SIZE = 100 * 1024 * 1024

EXT_SIZE_MAPPING = {
    ".ini": CONFIG_FILE_MAX_SIZE,
    '.csv': CSV_FILE_MAX_SIZE,
    '.json': JSON_FILE_MAX_SIZE,
    '.txt': TEXT_FILE_MAX_SIZE,
    '.py': TEXT_FILE_MAX_SIZE,
    '.pth': MODEL_WEIGHT_MAX_SIZE,
    '.bin': MODEL_WEIGHT_MAX_SIZE,
    '.onnx': ONNX_MODEL_MAX_SIZE,
}

MAX_RECUR_DEPTH = 998


class FileCheckConst:
    """
    Class for file check const
    """
    READ_ABLE = "r"
    WRITE_ABLE = "w"
    EXECUTE_ABLE = "x"
    READ_WRITE_ABLE = "rw"
    READ_EXECUTE_ABLE = "rx"
    WRITE_EXECUTE_ABLE = "wx"
    READ_WRITE_EXECUTE_ABLE = "rwx"
    PERM_OPTIONS = [READ_ABLE, WRITE_ABLE, EXECUTE_ABLE, READ_WRITE_ABLE, READ_EXECUTE_ABLE, READ_WRITE_EXECUTE_ABLE]

    DIRECTORY_LENGTH = 4096
    FILE_NAME_LENGTH = 255
    FILE_VALID_PATTERN = r"^[a-zA-Z0-9_.:/-]+$"
    FILE_PATTERN = r'^[a-zA-Z0-9_./-]+$'
    PKL_SUFFIX = ".pkl"
    NUMPY_SUFFIX = ".npy"
    JSON_SUFFIX = ".json"
    PT_SUFFIX = ".pt"
    BIN_SUFFIX = ".bin"
    CSV_SUFFIX = ".csv"
    XLSX_SUFFIX = ".xlsx"
    YAML_SUFFIX = ".yaml"
    IR_SUFFIX = ".ir"
    ZIP_SUFFIX = ".zip"
    SHELL_SUFFIX = ".sh"
    LOG_SUFFIX = ".log"
    ONNX_SUFFIX = '.onnx'
    OM_SUFFIX = '.om'
    MAX_PKL_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_NUMPY_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_JSON_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_PT_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_BIN_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_CSV_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_XLSX_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_YAML_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_IR_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_ZIP_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    MAX_FILE_IN_ZIP_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_FILE_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    COMMON_FILE_SIZE = 1073741824  # 1 * 1024 * 1024 * 1024
    MAX_LOG_SIZE = 10737418240  # 1 * 1024 * 1024 * 1024
    MAX_COMMON_FILE_SIZE = 10 * 1024 * 1024 * 1024  # 10GB
    MAX_ONNX_SIZE = 10737418240  # 1 * 1024 * 1024 * 1024
    MAX_OM_SIZE = 10737418240  # 10 * 1024 * 1024 * 1024
    DIR = "dir"
    FILE = "file"
    DATA_DIR_AUTHORITY = 0o750
    DATA_FILE_AUTHORITY = 0o640
    FILE_SIZE_DICT = {
        PKL_SUFFIX: MAX_PKL_SIZE,
        NUMPY_SUFFIX: MAX_NUMPY_SIZE,
        JSON_SUFFIX: MAX_JSON_SIZE,
        PT_SUFFIX: MAX_PT_SIZE,
        BIN_SUFFIX: MAX_BIN_SIZE,
        CSV_SUFFIX: MAX_CSV_SIZE,
        XLSX_SUFFIX: MAX_XLSX_SIZE,
        YAML_SUFFIX: MAX_YAML_SIZE,
        IR_SUFFIX: MAX_IR_SIZE,
        ZIP_SUFFIX: MAX_ZIP_SIZE,
        LOG_SUFFIX: MAX_LOG_SIZE,
        ONNX_SUFFIX: MAX_ONNX_SIZE,
        OM_SUFFIX: MAX_OM_SIZE
    }
    CSV_BLACK_LIST = r'^[＋－＝％＠\+\-=%@]|;[＋－＝％＠\+\-=%@]'



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import OrderedDict
from dataclasses import dataclass
import sys
import time
import psutil

from msprobe.core.common.file_utils import check_file_or_directory_path, load_json
from msprobe.core.common.const import Const


@dataclass
class RankPath:
    rank: int
    dump_path: str

    def __init__(self, rank, dump_path):
        self.rank = rank
        check_file_or_directory_path(dump_path)
        self.dump_path = dump_path


class FileCache:
    """
    lazy load file
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        self._max_memory_usage = psutil.virtual_memory().available / 4  # 最大占用当前可用内存空间的1/4
        self._cache = OrderedDict()
        self._access_cnt = {}
        self._access_time = {}
        self._size = {}

    @staticmethod
    def _sizeof(obj):
        seen = set()
        objs = [obj]
        size = 0
        while objs:
            obj = objs.pop()
            obj_id = id(obj)
            if obj_id in seen:
                continue
            seen.add(obj_id)
            size += sys.getsizeof(obj)
            if isinstance(obj, dict):
                objs.extend(obj.keys())
                objs.extend(obj.values())
            elif isinstance(obj, (list, tuple, set, frozenset)):
                objs.extend(obj)
        return size

    def load_json(self, json_path):
        if json_path in self._cache:
            self._access_cnt[json_path] += 1
            self._access_time[json_path] = time.monotonic()
            self._cache.move_to_end(json_path)
            return self._cache[json_path]
        self._cleanup()
        return self._load(json_path)

    def _load(self, json_path):
        data = load_json(json_path)
        self._add_to_cache(json_path, data)
        return data

    def _add_to_cache(self, key, data):
        if key in self._cache:
            self._cache.move_to_end(key)
        else:
            self._cache[key] = data
            self._access_cnt[key] = 0
            self._access_time[key] = time.monotonic()
            self._size[key] = self._sizeof(data)

    def _calc_cache_size(self):
        return sys.getsizeof(self._cache) + sum(self._size.values())

    def _cleanup(self):
        while self._calc_cache_size() > self._max_memory_usage and self._cache:
            least_frequent_key = min(self._access_cnt.keys(), key=lambda k: self._access_cnt[k])
            least_recent_key = min(self._access_time.keys(), key=lambda k: self._access_time[k])
            largest_key = max(self._cache.keys(), key=lambda k: self._size[k])
            key_to_rm = min([least_frequent_key, least_recent_key, largest_key],
                            key=lambda k: (self._access_cnt[k], self._access_time[k], -self._size[k]))
            del self._cache[key_to_rm]
            del self._access_cnt[key_to_rm]
            del self._access_time[key_to_rm]
            del self._size[key_to_rm]


def is_communication_op(op_name):
    # 定义通信算子的关键字，覆盖各种通信操作，如all_reduce, send, broadcast等
    # 从wrap文件中读取，先硬编码在文件中
    return (op_name.startswith((Const.DISTRIBUTED, Const.MINT_DIST_API_TYPE_PREFIX, Const.MS_API_TYPE_COM)) and
            any(keyword in op_name for keyword in DiffAnalyseConst.COMMUNICATION_KEYWORDS))


def is_ignore_op(op_name):
    ignore_keywords = [
        'Torch.empty',
        'Torch.fill',
        'Tensor.__setitem__'
    ]
    return any(keyword in op_name for keyword in ignore_keywords)


class DiffAnalyseConst:
    COMMUNICATION_KEYWORDS = {
        'send',  # send 算子
        'recv',  # recv 算子
        'broadcast',  # broadcast 算子
        'all_reduce',  # all_reduce 算子
        'reduce',  # reduce 算子
        'all_gather',  # all_gather 算子
        'gather',  # gather 算子
        'isend',  # isend 算子
        'irecv',  # irecv 算子
        'scatter',  # scatter 算子
        'reduce_scatter',  # reduce_scatter 算子
        '_reduce_scatter_base',  # _reduce_scatter_base 算子
        '_all_gather_base',  # _all_gather_base 算子
        'all_to_all_single',  # all_to_all_single 算子
        'all_to_all',  # all_to_all 算子
        'all_gather_into_tensor',  # all_gather_into_tensor 算子
        'reduce_scatter_tensor',  # reduce_scatter_tensor 算子
        'send_object_list',  # send_object_list 算子
        'recv_object_list'  # recv_object_list 算子
    }
    P2P_API_MAPPING = {'send': 'recv', 'recv': 'send', 'isend': 'irecv', 'irecv': 'isend',
                       'send_object_list': 'recv_object_list', 'recv_object_list': 'send_object_list'}
    SRC = 'src'
    DST = 'dst'
    SRC_GROUP = 'group_src'
    DST_GROUP = 'group_dst'
    LINK = 'link'
    DIRECTED_API = {'send': DST, 'recv': SRC, 'isend': DST, 'irecv': SRC, 'broadcast': SRC, 'scatter': SRC,
                    'gather': DST, 'send_object_list': DST, 'recv_object_list': SRC}
    OPPOSITE_DIR = {SRC: DST, DST: SRC}
    DUMP_FILE = "dump.json"
    CONSTRUCT_FILE = "construct.json"
    STACK_FILE = "stack.json"


def analyze_diff_in_group(nodes_group):
    diff_nodes = []

    def get_compute_ops_from_comm_nodes(comm_nodes):
        for comm_node in comm_nodes:
            for op_node in comm_node.compute_ops:
                op_node.layer = comm_node.layer
                diff_nodes.append(op_node)

    def get_comm_ops(comm_nodes):
        for node in comm_nodes:
            node.data.layer = node.layer
            diff_nodes.append(node.data)

    # 先看src或link中input是否有异常
    src_list = list(filter(lambda node: node.type in [DiffAnalyseConst.SRC, DiffAnalyseConst.LINK], nodes_group))
    input_diff_nodes = list(filter(lambda node: node.is_diff, src_list))
    # 如果有异常回溯计算节点找到异常来源
    # 使用cpu模拟节点进行计算，查看结果是否有问题。需要对所有计算节点录入/映射，暂不实现。
    get_compute_ops_from_comm_nodes(nodes_group)
    # 筛选入参没问题但出参有问题的通信节点
    output_diff_nodes = list(filter(lambda node: node.data.is_diff, nodes_group))
    get_comm_ops(output_diff_nodes)
    return diff_nodes


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.log import logger
from msprobe.core.common.utils import check_op_str_pattern_valid, CompareException
from msprobe.core.common.const import Const

cross_dtype_mapping = {
    "Int8": "int",
    "torch.int8": "int",
    "UInt8": "int",
    "torch.uint8": "int",
    "Int16": "int",
    "torch.int16": "int",
    "UInt16": "int",
    "torch.uint16": "int",
    "Int32": "int",
    "torch.int32": "int",
    "UInt32": "int",
    "torch.uint32": "int",
    "Int64": "int",
    "torch.int64": "int",
    "UInt64": "int",
    "torch.uint64": "int",

    "Float16": "float",
    "torch.float16": "float",
    "Float32": "float",
    "torch.float32": "float",
    "Float64": "float",
    "torch.float64": "float",
    "BFloat16": "float",
    "torch.bfloat16": "float",

    "Bool": "bool",
    "torch.bool": "bool",

    "Complex64": "complex",
    "torch.complex64": "complex",
    "Complex128": "complex",
    "torch.complex128": "complex",
}


def check_dump_json_str(op_data, op_name):
    input_list = op_data.get(Const.INPUT_ARGS, None) if op_data.get(Const.INPUT_ARGS, None) else op_data.get(
        Const.INPUT, None)
    input_kwargs = op_data.get(Const.INPUT_KWARGS, None)
    output_list = op_data.get(Const.OUTPUT, None)

    args = [input_list, input_kwargs, output_list]
    for arg in args:
        if not arg:
            continue
        if isinstance(arg, dict):
            check_json_key_value(arg, op_name)
        else:
            for ele in arg:
                if not ele:
                    continue
                check_json_key_value(ele, op_name)


def check_json_key_value(input_output, op_name, depth=0):
    if depth > Const.MAX_DEPTH:
        logger.error(f"string check of data info of {op_name} exceeds the recursion limit.")
        return
    if isinstance(input_output, list):
        for item in input_output:
            check_json_key_value(item, op_name, depth + 1)
    elif isinstance(input_output, dict):
        for key, value in input_output.items():
            if isinstance(value, dict):
                check_json_key_value(value, op_name, depth + 1)
            else:
                valid_key_value(key, value, op_name)


def valid_key_value(key, value, op_name):
    if key == "shape" and not isinstance(value, (list, tuple)):
        logger.error(f"shape of input or output of {op_name} is not list or tuple, please check!")
        raise CompareException(CompareException.INVALID_OBJECT_TYPE_ERROR)
    elif key == "requires_grad" and not isinstance(value, bool):
        logger.error(f"requires_grad of input or output of {op_name} is not bool, please check!")
        raise CompareException(CompareException.INVALID_OBJECT_TYPE_ERROR)
    else:
        check_op_str_pattern_valid(value, op_name)


def check_stack_json_str(stack_info, op_name):
    if isinstance(stack_info, list):
        for item in stack_info:
            check_op_str_pattern_valid(item, op_name, stack=True)
    else:
        logger.error(f"Expected stack_info to be a list, but got {type(stack_info).__name__} for '{op_name}'")
        raise CompareException(CompareException.INVALID_OBJECT_TYPE_ERROR)


def check_configuration_param(config):
    arg_list = [config.stack_mode, config.fuzzy_match, config.first_diff_analyze, config.is_print_compare_log]
    arg_names = ['stack_mode', 'fuzzy_match', 'first_diff_analyze', 'is_print_compare_log']
    for arg, name in zip(arg_list, arg_names):
        if not isinstance(arg, bool):
            logger.error(f"Invalid input parameter, {name} which should be only bool type.")
            raise CompareException(CompareException.INVALID_PARAM_ERROR)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch
from typing import Callable
from msprobe.pytorch.api_accuracy_checker.common.utils import is_dtype_fp8_or_hif8
from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import absolute_standard_api, binary_standard_api, \
    ulp_standard_api, thousandth_standard_api, accumulative_error_standard_api, BINARY_COMPARE_UNSUPPORT_LIST
from msprobe.core.common.const import CompareConst


class StandardRegistry:
    """
    Registry class for managing comparison standards and functions.

    This class provides a centralized registry for different comparison standards and their corresponding functions.
    It allows for dynamic registration of comparison functions based on the standard category.

    Attributes:
        comparison_functions (dict): A dictionary mapping standard categories to their corresponding comparison 
        functions.
        standard_categories (dict): A dictionary mapping standard names to their corresponding API categories.

    Methods:
        _get_standard_category(api_name, dtype): Determines the standard category for a given API name and data type.
        register(standard, func): Registers a comparison function for a given standard category.
        get_comparison_function(api_name, dtype): Retrieves the comparison function for a given API name and data type.

    Note:
        The data type is used to determine the standard category if it is not supported by binary comparison.
        If the API name is not found in any standard category, it defaults to the 'benchmark' category.

    See Also:
        BaseCompare: The base class for comparison classes.
    """
    def __init__(self):
        self.comparison_functions = {}
        self.api_standard_function_map = {
            CompareConst.ABSOLUTE_THRESHOLD: absolute_standard_api,
            CompareConst.BINARY_CONSISTENCY: binary_standard_api,
            CompareConst.ULP_COMPARE: ulp_standard_api,
            CompareConst.THOUSANDTH_STANDARD: thousandth_standard_api,
            CompareConst.ACCUMULATIVE_ERROR_COMPARE: accumulative_error_standard_api
        }

    def register(self, standard: str, func: Callable) -> None:
        """
        Registers a comparison function for a given standard category.

        Args:
            standard (str): The name of the standard category.
            func (Callable): The comparison function to be registered.

        Raises:
            ValueError: If the standard category is not supported.
        """
        if not callable(func):
            raise ValueError("The function to be registered must be callable.")
        self.comparison_functions[standard] = func

    def get_comparison_function(self, api_name, dtype, in_dtype):
        standard = self._get_standard_category(api_name, dtype, in_dtype)
        return self.comparison_functions.get(standard)

    def _get_standard_category(self, api_name, out_dtype, in_dtype):
        """
        Determines the standard category for a given API name and data type.

        This method checks if the provided data type is supported for binary comparison.
        If it is, the method returns 'binary_consistency'. Otherwise, it iterates over the
        api_standard_function_map to find a matching category for the API name.

        Args:
            api_name (str): The name of the API for which to determine the standard category.
            out_dtype (type): The data type of the output tensor.
            in_dtype (type): The data type of the input tensor.

        Returns:
            str: The name of the standard category that matches the API name and data type, or 'benchmark' if no match 
            is found.

        Note:
        This method assumes that the api_standard_function_map is properly populated with standard categories and 
        their corresponding API functions.
        The BINARY_COMPARE_UNSUPPORT_LIST should be defined and contain all data types that are not supported for 
        binary comparison.
        """
        if is_dtype_fp8_or_hif8(out_dtype):
            return CompareConst.ULP_COMPARE
        if is_dtype_fp8_or_hif8(in_dtype):
            return CompareConst.ABSOLUTE_THRESHOLD
        if out_dtype in ("torch.float16", "torch.bfloat16", "torch.float32", "torch.float64",
        torch.float16, torch.bfloat16, torch.float32, torch.float64):
            return CompareConst.BENCHMARK  
        if str(out_dtype) not in BINARY_COMPARE_UNSUPPORT_LIST:
            return CompareConst.BINARY_CONSISTENCY
        for name, category in self.api_standard_function_map.items():
            if api_name in category:
                return name
        return CompareConst.BENCHMARK



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import os
import re
import sys
import time
import gc
from collections import namedtuple

try:
    import torch_npu
except ImportError:
    is_gpu = True
    current_device = "cuda"
else:
    is_gpu = False
    current_device = "npu"

import torch
from tqdm import tqdm

from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import BackwardMessage, UtDataInfo, \
    get_validated_result_csv_path, get_validated_details_csv_path, exec_api, record_skip_info, is_unsupported_api
from msprobe.pytorch.api_accuracy_checker.acc_check.data_generate import gen_api_params, gen_args
from msprobe.pytorch.api_accuracy_checker.common.utils import api_info_preprocess, \
    initialize_save_path, UtDataProcessor, extract_basic_api_segments, ApiData
from msprobe.pytorch.api_accuracy_checker.compare.compare import Comparator
from msprobe.pytorch.api_accuracy_checker.compare.compare_column import CompareColumn
from msprobe.pytorch.api_accuracy_checker.common.config import CheckerConfig
from msprobe.pytorch.common.parse_json import parse_json_info_forward_backward
from msprobe.core.common.file_utils import FileChecker, change_mode, \
    create_directory, get_json_contents, read_csv, check_file_or_directory_path
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.dump.pt_config import parse_json_config
from msprobe.core.common.const import Const, FileCheckConst, CompareConst
from msprobe.core.common.utils import safe_get_value, CompareException, is_int, check_op_str_pattern_valid
from msprobe.pytorch.common.utils import seed_all
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import generate_cpu_params,\
    generate_device_params, \
    ExecParams


current_time = time.strftime("%Y%m%d%H%M%S")
UT_ERROR_DATA_DIR = 'ut_error_data' + current_time
RESULT_FILE_NAME = "accuracy_checking_result_" + current_time + ".csv"
DETAILS_FILE_NAME = "accuracy_checking_details_" + current_time + ".csv"


not_backward_list = ['repeat_interleave']
unsupported_backward_list = ['masked_select']
unsupported_api_list = ["to", "empty", "empty_like", "empty_strided", "new_empty", "new_empty_strided", 
                        "empty_with_format"]


tqdm_params = {
    'smoothing': 0,  # 平滑进度条的预计剩余时间，取值范围0到1
    'desc': 'Processing',  # 进度条前的描述文字
    'leave': True,  # 迭代完成后保留进度条的显示
    'ncols': 75,  # 进度条的固定宽度
    'mininterval': 0.1,  # 更新进度条的最小间隔秒数
    'maxinterval': 1.0,  # 更新进度条的最大间隔秒数
    'miniters': 1,  # 更新进度条之间的最小迭代次数
    'ascii': None,  # 根据环境自动使用ASCII或Unicode字符
    'unit': 'it',  # 迭代单位
    'unit_scale': True,  # 自动根据单位缩放
    'dynamic_ncols': True,  # 动态调整进度条宽度以适应控制台
    'bar_format': '{l_bar}{bar}| {n}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}]'  # 自定义进度条输出格式
}


seed_all()


def acc_check(config):
    logger.info("start acc_check test")

    logger.info(f"acc_check task result will be saved in {config.result_csv_path}")
    logger.info(f"acc_check task details will be saved in {config.details_csv_path}")

    if config.save_error_data:
        logger.info(f"acc_check task error_data will be saved in {config.error_data_path}")
    compare = Comparator(config.result_csv_path, config.details_csv_path, config.is_continue_acc_check, config=config)


    csv_df = read_csv(config.result_csv_path)
    try:
        api_name_set = {row[0] for row in csv_df.itertuples(index=False, name=None)}
    except IndexError:
        logger.error(f"Read {config.result_csv_path} error, api_name_set is empty.")
        api_name_set = set()
    run_api_offline(config, compare, api_name_set)
    for result_csv_path, details_csv_path in zip(compare.save_path_list, compare.detail_save_path_list):
        change_mode(result_csv_path, FileCheckConst.DATA_FILE_AUTHORITY)
        change_mode(details_csv_path, FileCheckConst.DATA_FILE_AUTHORITY)
        logger.info(f"acc_check task result csv is saved in {result_csv_path}")
        logger.info(f"acc_check task details csv is saved in {details_csv_path}")
    compare.print_pretest_result()


def run_api_offline(config, compare, api_name_set):
    err_column = CompareColumn()
    for _, (api_full_name, api_info_dict) in enumerate(tqdm(config.forward_content.items(), **tqdm_params)):
        check_op_str_pattern_valid(api_full_name)
        if api_full_name in api_name_set:
            continue
        if is_unsupported_api(api_full_name):
            skip_message = f"API {api_full_name} not support for acc_check. SKIP."
            compare_alg_results = err_column.to_column_value(CompareConst.SKIP, skip_message)
            record_skip_info(api_full_name, compare, compare_alg_results)
            continue
        _, api_name = extract_basic_api_segments(api_full_name)
        if not api_name:
            err_message = f"API {api_full_name} not support for acc_check. SKIP."
            logger.error(err_message)
            compare_alg_results = err_column.to_column_value(CompareConst.SKIP, err_message)
            record_skip_info(api_full_name, compare, compare_alg_results)
            continue
        try:
            if blacklist_and_whitelist_filter(api_name, config.black_list, config.white_list):
                skip_message = f"API {api_name} in black list or not in white list. SKIP."
                logger.info(skip_message)
                compare_alg_results = err_column.to_column_value(CompareConst.SKIP, skip_message)
                record_skip_info(api_full_name, compare, compare_alg_results)
                continue
            data_info = run_torch_api(api_full_name, config.real_data_path, config.backward_content, api_info_dict)
            is_fwd_success, is_bwd_success = compare.compare_output(api_full_name, data_info)
            if config.save_error_data:
                do_save_error_data(api_full_name, data_info, config.error_data_path, is_fwd_success, is_bwd_success)
        except Exception as err:
            if "expected scalar type Long" in str(err):
                logger.warning(f"API {api_name} not support int32 tensor in CPU, please add {api_name} to CONVERT_API "
                               "'int32_to_int64' list in accuracy_tools/msprobe/core/common/const.py file.")
            else:
                logger.error(f"Run {api_full_name} acc_check Error: %s" % str(err))
            compare_alg_results = err_column.to_column_value(CompareConst.SKIP, str(err))
            record_skip_info(api_full_name, compare, compare_alg_results)
        finally:
            if is_gpu:
                torch.cuda.empty_cache()
            else:
                torch.npu.empty_cache()
            gc.collect()


def blacklist_and_whitelist_filter(api_name, black_list, white_list):
    """
    run api(api_name) if api_name not in black_list and in white_list.
    If api is both in black_list and black_list, black_list first.
    return: False for exec api, True for not exec
    """
    black_list.extend(unsupported_api_list)
    if black_list and api_name in black_list:
        return True
    if white_list and api_name not in white_list:
        return True
    return False


def do_save_error_data(api_full_name, data_info, error_data_path, is_fwd_success, is_bwd_success):
    if not is_fwd_success or not is_bwd_success:
        processor = UtDataProcessor(error_data_path)
        for element in data_info.in_fwd_data_list:
            processor.save_tensors_in_element(api_full_name + '.forward.input', element)
        processor.save_tensors_in_element(api_full_name + '.forward.output.bench', data_info.bench_output)
        processor.save_tensors_in_element(api_full_name + '.forward.output.device', data_info.device_output)
        processor.save_tensors_in_element(api_full_name + '.backward.input', data_info.grad_in)
        processor.save_tensors_in_element(api_full_name + '.backward.output.bench', data_info.bench_grad)
        processor.save_tensors_in_element(api_full_name + '.backward.output.device', data_info.device_grad)


def run_torch_api(api_full_name, real_data_path, backward_content, api_info_dict):
    in_fwd_data_list = []
    backward_message = ''
    api_type, api_name = extract_basic_api_segments(api_full_name)
    args, kwargs, output_dtype = get_api_info(api_info_dict, api_name, real_data_path)
    need_grad = check_need_grad(api_info_dict)
    in_fwd_data_list.append(args)
    in_fwd_data_list.append(kwargs)
    need_backward = api_full_name in backward_content
    if not need_grad:
        logger.warning("%s %s" % (api_full_name, BackwardMessage.UNSUPPORT_BACKWARD_MESSAGE))
        backward_message += BackwardMessage.UNSUPPORT_BACKWARD_MESSAGE
    if api_name in not_backward_list:
        need_grad = False
        logger.info("%s %s" % (api_full_name, BackwardMessage.NO_BACKWARD_RESULT_MESSAGE))
        backward_message += BackwardMessage.NO_BACKWARD_RESULT_MESSAGE
    if api_name in unsupported_backward_list:
        need_grad = False
        logger.info("%s %s" % (api_full_name, BackwardMessage.UNSUPPORT_API_MESSAGE))
        backward_message += BackwardMessage.UNSUPPORT_API_MESSAGE
    need_backward = need_backward and need_grad
    
    device_info_kwargs = kwargs.get(Const.DEVICE)
    if device_info_kwargs and device_info_kwargs.get(Const.VALUE):
        kwargs[Const.DEVICE] = current_device
    device_args, device_kwargs, is_fp8 = generate_device_params(args, kwargs, need_backward, api_name)
    if kwargs.get(Const.DEVICE):
        del kwargs[Const.DEVICE]
    cpu_params = generate_cpu_params(args, kwargs, need_backward, api_name)
    cpu_args, cpu_kwargs = cpu_params.cpu_args, cpu_params.cpu_kwargs
    autocast_dtype, is_autocast = cpu_params.autocast_dtype, cpu_params.is_autocast
    if not is_autocast and output_dtype:
        is_autocast = autocast_dtype != output_dtype
        autocast_dtype = output_dtype
    bench_grad_out, device_grad_out = None, None
    cpu_exec_params = ExecParams(api_type, api_name, Const.CPU_LOWERCASE, cpu_args, cpu_kwargs, False, autocast_dtype)
    out = exec_api(cpu_exec_params)
    device_exec_params = ExecParams(api_type, api_name, current_device, device_args, device_kwargs, is_autocast,
                                     autocast_dtype)
    device_out = exec_api(device_exec_params)
    if is_fp8 and isinstance(device_out, torch.Tensor) and device_out.dtype == torch.float32:
        device_out = device_out.to(torch.float16)
    current_path = os.path.dirname(os.path.realpath(__file__))
    ut_setting_path = os.path.join(current_path, "torch_ut_setting.json")
    api_setting_dict = get_json_contents(ut_setting_path)
    grad_input_index = api_setting_dict.get(api_name)
    grad_index = None
    grad, bench_grad = None, None
    if grad_input_index is not None:
        grad_index = grad_input_index.get('grad_index')

    if need_backward and out is not None:
        if need_to_backward(grad_index, out):
            backward_args = backward_content[api_full_name].get("input")
            func_options = {
                'real_data_path': real_data_path
            }
            grad = gen_args(backward_args, api_name, func_options)
            grad = safe_get_value(grad, 0, "grad")
            grad_params = generate_cpu_params(grad, {}, False, api_name)
            bench_grad = grad_params.cpu_args
            bench_grad_out = run_backward(cpu_args, bench_grad, grad_index, out)
            device_grad = grad.clone().detach().to(current_device)
            device_grad_out = run_backward(device_args, device_grad, grad_index, device_out)
        else:
            backward_message += BackwardMessage.MULTIPLE_BACKWARD_MESSAGE
    if api_name == "npu_fusion_attention":
        out = safe_get_value(out, 0, "out")
        device_out = safe_get_value(device_out, 0, "device_out")

    return UtDataInfo(bench_grad_out, device_grad_out, device_out, out, bench_grad, in_fwd_data_list, backward_message,
                      rank=0, is_fp8=is_fp8)


def check_need_grad(api_info_dict):
    need_grad = True
    if api_info_dict.get(Const.INPUT_KWARGS) and "out" in api_info_dict.get(Const.INPUT_KWARGS):
        need_grad = False
    return need_grad


def get_api_info(api_info_dict, api_name, real_data_path):
    convert_type, api_info_dict = api_info_preprocess(api_name, api_info_dict)
    need_grad = check_need_grad(api_info_dict)
    args, kwargs, output_dtype = gen_api_params(api_info_dict, api_name, need_grad, convert_type, real_data_path)
    return args, kwargs, output_dtype


def need_to_backward(grad_index, out):
    if grad_index is None and isinstance(out, (list, tuple)):
        return False
    return True


def run_backward(args, grad, grad_index, out):
    if grad_index is not None:
        if not is_int(grad_index):
            logger.error(f"{grad_index} dtype is not int")
            raise TypeError(f"{grad_index} dtype is not int")
        if grad_index >= len(out):
            logger.error(f"Run backward error when grad_index is {grad_index}")
            raise IndexError(f"Run backward error when grad_index is {grad_index}")
        out[grad_index].backward(grad)
    else:
        out.backward(grad)

    grad_out = extract_tensors_grad(args)

    return grad_out


def extract_tensors_grad(args, depth=0):
    if depth > Const.MAX_DEPTH:
        logger.error("The depth of arg_in is too large, please check the arg_in.")
        raise CompareException(CompareException.RECURSION_LIMIT_ERROR)
    grads = []
    for arg in args:
        if isinstance(arg, torch.Tensor):
            grads.append(arg.grad)
        elif isinstance(arg, (list, tuple)):
            grads.extend(extract_tensors_grad(arg, depth + 1))
    return grads


def initialize_save_error_data(error_data_path):
    create_directory(error_data_path)
    error_data_path_checker = FileChecker(error_data_path, FileCheckConst.DIR,
                                          ability=FileCheckConst.WRITE_ABLE)
    error_data_path = error_data_path_checker.common_check()
    error_data_path = initialize_save_path(error_data_path, UT_ERROR_DATA_DIR)
    return error_data_path


def _acc_check_parser(parser):
    parser.add_argument("-api_info", "--api_info_file", dest="api_info_file", default="", type=str,
                        help="<Optional> The api param tool result file: generate from api param tool, "
                             "a json file.",
                        required=False)
    parser.add_argument("-o", "--out_path", dest="out_path", default="", type=str,
                        help="<optional> The acc_check task result out path.",
                        required=False)
    parser.add_argument('-save_error_data', dest="save_error_data", action="store_true",
                        help="<optional> Save compare failed api output.", required=False)
    parser.add_argument("-j", "--jit_compile", dest="jit_compile", action="store_true",
                        help="<optional> whether to turn on jit compile", required=False)

    class UniqueDeviceAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            unique_values = set(values)
            if len(values) != len(unique_values):
                parser.error("device id must be unique")
            for device_id in values:
                if not 0 <= device_id:
                    parser.error("device id must be greater than or equal to 0")
            setattr(namespace, self.dest, values)

    parser.add_argument("-d", "--device", dest="device_id", nargs='+', type=int,
                        help="<optional> set device id to acc_check, must be unique and in range 0-7",
                        default=[0], required=False, action=UniqueDeviceAction)
    parser.add_argument("-csv_path", "--result_csv_path", dest="result_csv_path", default="", type=str,
                        help="<optional> The path of accuracy_checking_result_{timestamp}.csv, "
                             "when acc_check is interrupted, enter the file path to continue acc_check.",
                        required=False)
    parser.add_argument("-f", "--filter_api", dest="filter_api", action="store_true",
                        help="<optional> Whether to filter the api in the api_info_file.", required=False)
    parser.add_argument("-config", "--config_path", dest="config_path", default="", type=str,
                        help="<optional> The path of config.json", required=False)


def preprocess_forward_content(forward_content):
    processed_content = {}
    base_keys_variants = {}
    arg_cache = {}

    for key, value in forward_content.items():
        check_op_str_pattern_valid(key)
        base_key = key.rsplit(Const.SEP, 1)[0]

        if key not in arg_cache:
            filtered_new_args = [
                {k: v for k, v in arg.items() if k not in ['Max', 'Min']}
                for arg in value['input_args']
                if isinstance(arg, dict)
            ]
            arg_cache[key] = (filtered_new_args, value['input_kwargs'])

        filtered_new_args, new_kwargs = arg_cache[key]

        if base_key not in base_keys_variants:
            processed_content[key] = value
            base_keys_variants[base_key] = {key}
        else:
            is_duplicate = False
            for variant in base_keys_variants.get(base_key, []):
                try:
                    existing_args, existing_kwargs = arg_cache.get(variant)
                except KeyError as e:
                    logger.error(f"KeyError: {e} when processing {key}")
                if existing_args == filtered_new_args and existing_kwargs == new_kwargs:
                    is_duplicate = True
                    break

            if not is_duplicate:
                processed_content[key] = value
                base_keys_variants[base_key].add(key)

    return processed_content


def _acc_check(parser=None):
    if not parser:
        parser = argparse.ArgumentParser()
    _acc_check_parser(parser)
    args = parser.parse_args(sys.argv[1:])
    acc_check_command(args)


def acc_check_command(args):
    if args.config_path:
        config_path_checker = FileChecker(args.config_path, FileCheckConst.FILE, 
                                          FileCheckConst.READ_ABLE, FileCheckConst.JSON_SUFFIX)
        checked_config_path = config_path_checker.common_check()
        _, task_config = parse_json_config(checked_config_path, Const.ACC_CHECK)
        checker_config = CheckerConfig(task_config)
    else:
        checker_config = CheckerConfig()
    
    if not args.api_info_file:
        logger.error("Please provide api_info_file for offline acc_check.")
        raise Exception("Please provide api_info_file for offline acc_check.")

    if not is_gpu:
        torch.npu.set_compile_mode(jit_compile=args.jit_compile)
        if args.jit_compile:
            torch.npu.config.allow_internal_format = True
        else:
            torch.npu.config.allow_internal_format = False
    used_device = current_device + ":" + str(args.device_id[0])
    try:
        if is_gpu:
            torch.cuda.set_device(used_device)
        else:
            torch.npu.set_device(used_device)
    except Exception as error:
        logger.error(f"Set device id failed. device id is: {args.device_id}")
        raise NotImplementedError from error

    # 在线预检场景下，不需要外出输出api信息，forward_content, backward_content, real_data_path设置为None
    # 离线场景下，forward_content, backward_content, real_data_path从api_info_file中解析
    forward_content, backward_content, real_data_path = None, None, None
    if args.api_info_file:
        api_info_file_checker = FileChecker(file_path=args.api_info_file, path_type=FileCheckConst.FILE, 
                                            ability=FileCheckConst.READ_ABLE, file_type=FileCheckConst.JSON_SUFFIX)
        checked_api_info = api_info_file_checker.common_check()
        forward_content, backward_content, real_data_path = parse_json_info_forward_backward(checked_api_info)
        if real_data_path:
            dump_path = os.path.dirname(checked_api_info)
            real_data_path = os.path.join(dump_path, Const.DUMP_TENSOR_DATA)
        if args.filter_api:
            logger.info("Start filtering the api in the api_info_file.")
            forward_content = preprocess_forward_content(forward_content)
            logger.info("Finish filtering the api in the api_info_file.")

    out_path = args.out_path if args.out_path else Const.DEFAULT_PATH
    create_directory(out_path)
    out_path_checker = FileChecker(out_path, FileCheckConst.DIR, ability=FileCheckConst.WRITE_ABLE)
    out_path = out_path_checker.common_check()
    save_error_data = args.save_error_data

    result_csv_path = os.path.join(out_path, RESULT_FILE_NAME)
    details_csv_path = os.path.join(out_path, DETAILS_FILE_NAME)
    if args.result_csv_path:
        result_csv_path = get_validated_result_csv_path(args.result_csv_path, 'result')
        details_csv_path = get_validated_details_csv_path(result_csv_path)

    error_data_path = checker_config.error_data_path
    if save_error_data:
        if args.result_csv_path:
            parts_by_dot = result_csv_path.split(Const.SEP)
            if len(parts_by_dot) < 2 or not parts_by_dot[0]:
                raise ValueError("result_csv_path does not contain a valid file name with an extension.")
            file_name_part = parts_by_dot[0]
            parts_by_underscore = file_name_part.split(Const.REPLACEMENT_CHARACTER)
            if len(parts_by_underscore) < 2:
                raise ValueError("File name part does not contain enough '_' separated segments.")
            time_info = parts_by_underscore[-1]

            global UT_ERROR_DATA_DIR
            UT_ERROR_DATA_DIR = 'ut_error_data' + time_info
        error_data_path = initialize_save_error_data(error_data_path)
    config_params = {
        'forward_content': forward_content,
        'backward_content': backward_content,
        'result_csv_path': result_csv_path,
        'details_csv_path': details_csv_path,
        'save_error_data': save_error_data,
        'is_continue_acc_check': args.result_csv_path,
        'real_data_path': real_data_path,
        'error_data_path': error_data_path
    }
    acc_check_config = checker_config.get_acc_check_config(**config_params)
    acc_check(acc_check_config)
    logger.info("acc_check task completed.")


if __name__ == '__main__':
    _acc_check()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from .graph_service_db import DbGraphService
from .graph_service_vis import JsonGraphService
from ..utils.global_state import GraphState
from ..utils.constant import DataType


class ServiceFactory:

    def __init__(self):
        self.run = ''
        self.tag = ''
        self.data_type = None
        self.strategy = JsonGraphService('', '')

    def create_strategy(self, data_type, run, tag):
        if not (data_type == self.data_type and run == self.run and tag == self.tag):
            if data_type == DataType.DB.value:
                self.strategy = DbGraphService(run, tag)
            else:
                self.strategy = JsonGraphService(run, tag)
            # 创建成功后，更新全局变量
            self.data_type = data_type
            self.run = run
            self.tag = tag
        return self.strategy
    
    def create_strategy_without_tag(self, data_type, run):
        if not (data_type == self.data_type and run == self.run):
   
            self.tag = GraphState.get_global_value('first_run_tags', {}).get(self.run)
            if data_type == DataType.DB.value:
                self.strategy = DbGraphService(run, self.tag)
            else:
                self.strategy = JsonGraphService(run, self.tag)
            # 创建成功后，更新全局变量
            self.data_type = data_type
            self.run = run
        return self.strategy



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
from collections import defaultdict
from functools import wraps
import os
import pickle
import re

from msprobe.infer.utils.constants import TENSOR_MAX_SIZE, EXT_SIZE_MAPPING, PATH_WHITE_LIST_REGEX, MAX_RECUR_DEPTH
from msprobe.core.common.log import logger
from msprobe.infer.utils.file_open_check import is_legal_path_length

# 记录工具函数递归的深度
recursion_depth = defaultdict(int)


def confirmation_interaction(prompt):
    confirm_pattern = re.compile(r'y(?:es)?', re.IGNORECASE)

    try:
        user_action = input(prompt)
    except Exception:
        return False

    return bool(confirm_pattern.match(user_action))


def check_file_ext(path, ext: str):
    if not isinstance(path, str):
        raise TypeError(f"Expected first positional argument type 'str', got {type(path)} instead")

    if not isinstance(ext, str):
        raise TypeError(f"Expected second positional argument type 'str', got {type(ext)} instead")

    path_ext = os.path.splitext(path)[1]

    if path_ext != ext:
        return False

    return True


def check_file_size_based_on_ext(path, ext=None):
    """Check the file size based on extension. This function uses `os.stat` to get file size may lead to OSError"""

    if not isinstance(path, str):
        raise TypeError(f"Expected path to be 'str', got {type(path)} instead")

    ext = ext or os.path.splitext(path)[1]
    size = os.path.getsize(path)  # may lead to errors

    if ext in EXT_SIZE_MAPPING:
        if size > EXT_SIZE_MAPPING[ext]:
            return False
    else:
        if size > TENSOR_MAX_SIZE:
            confirmation_prompt = "The file %r is larger than expected. " \
                                "Attempting to read such a file could potentially impact system performance.\n" \
                                "Please confirm your awareness of the risks associated with this action (y/n): " % path
            return confirmation_interaction(confirmation_prompt)

    return True


def safe_torch_load(path, **kwargs):
    import torch  # Do not move it !!! it may caused Import Error
    kwargs['weights_only'] = True
    tensor = None

    while True:
        try:
            tensor = torch.load(path, **kwargs)
        except pickle.UnpicklingError:
            confirmation_prompt = "Weights only load failed. Re-running `torch.load` with `weights_only` " \
                                  "set to `False` will likely succeed, but it can result in arbitrary code " \
                                  "execution. Do it only if you get the file from a trusted source.\n" \
                                  "Please confirm your awareness of the risks associated with this action ([y]/n): "
            if not confirmation_interaction(confirmation_prompt):
                raise
            kwargs['weights_only'] = False
        else:
            break

    return tensor


def load_file_to_read_common_check(path: str, exts=None):
    if not isinstance(path, str):
        raise TypeError("'path' should be 'str'")

    if isinstance(exts, (tuple, list)):
        if not any(check_file_ext(path, ext) for ext in exts):
            logger.error(f"Expected extension to be one of {exts}")
            raise ValueError

    elif exts is not None:
        logger.error(f"Expected 'exts' to be 'List[str]', got {type(exts)} instead")
        raise TypeError

    if re.search(PATH_WHITE_LIST_REGEX, path):
        logger.error(f"Invalid character: {path}")
        raise ValueError

    if not is_legal_path_length(path):
        logger.error("Invalid path length.")
        raise ValueError

    path = os.path.realpath(path)

    try:
        file_status = os.stat(path)
    except OSError as e:
        logger.error(f"{e.strerror}: {path}")
        raise

    if not os.st.S_ISREG(file_status.st_mode):
        logger.error(f"Not a regular file: {path}")
        raise ValueError

    if not check_file_size_based_on_ext(path):
        logger.error(f"File too large: {path}")
        raise ValueError

    if (os.st.S_IWOTH & file_status.st_mode) == os.st.S_IWOTH:
        logger.error(f"Vulnerable path: {path} should not be other writeable")
        raise PermissionError

    cur_euid = os.geteuid()
    if file_status.st_uid != cur_euid:
        # not root
        if cur_euid != 0:
            logger.error(f"File owner and current user are inconsistent: {path}")
            raise PermissionError

        # root but reading a other writeable file
        elif (os.st.S_IWGRP & file_status.st_mode) == os.st.S_IWGRP or \
             (os.st.S_IWUSR & file_status.st_mode) == os.st.S_IWUSR:
            logger.warning("Privilege escalation risk detected. Trying to read a file that belongs to"
                           " a normal user and is writeable to the user or the user group")

    return path


def is_valid_command(arg_str, index):
    first_whitelist_pattern = re.compile(r"^[a-zA-Z0-9_\-./=:,\[\] ]+$")
    whitelist_pattern = re.compile(r"^[a-zA-Z0-9_\-./=:,\[\] ;]+$")

    if index == 0:
        return re.fullmatch(first_whitelist_pattern, arg_str), first_whitelist_pattern
    else:
        return re.fullmatch(whitelist_pattern, arg_str), whitelist_pattern


def filter_cmd(paras):
    filtered = []
    for index, arg in enumerate(paras):
        arg_str = str(arg)
        valid, pattern = is_valid_command(arg_str, index)
        if not valid:
            raise ValueError(f"The command contains invalid characters. Only the '{pattern}' pattern is allowed.")
        filtered.append(arg_str)
    return filtered



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the node advisor function.
"""

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.advisor.advisor_const import AdvisorConst
from msprobe.msaccucmp.advisor.advisor_result import AdvisorResult


class NodeAdvisor:
    """
    Class for generate node advisor
    """

    def __init__(self, input_file, result):
        self.analyze_data = input_file
        self.result = result

    def start_analyze(self):
        """
        Analyze result by node detection
        """
        log.print_info_log('Start Global Consistency detection.')
        data_columns = self.analyze_data.columns.values
        if AdvisorConst.COSINE_SIMILARITY not in data_columns:
            log.print_warn_log('Input csv file does not contain %s columns, Skip Global Consistency detection.'
                               % AdvisorConst.COSINE_SIMILARITY)
            return self.result
        else:
            have_cos_df = self.analyze_data.dropna(subset=[AdvisorConst.COSINE_SIMILARITY])
            # check cosine dataframe lines
            if have_cos_df.shape[0] == 0:
                log.print_warn_log('After analysis, input csv file %s column, does not have valid value. '
                                   'May all values be NAN, please check.'
                                   % AdvisorConst.COSINE_SIMILARITY)
                return self.result
            cos_df = have_cos_df.reset_index(drop=True, inplace=False)
            err_cos_df = have_cos_df[have_cos_df[AdvisorConst.COSINE_SIMILARITY] < AdvisorConst.ACCURACY_THRESHOLD]
            if err_cos_df.shape[0] == 0:
                return AdvisorResult(True, AdvisorConst.CONSISTENCY_DETECTION,
                                     AdvisorConst.NO_ERROR_OP, AdvisorConst.CONSISTENCY_SUGGEST)
            err_cos_df.reset_index(drop=True, inplace=True)
            index = err_cos_df.at[0, AdvisorConst.INDEX]
            cos_df_len = cos_df.shape[0]
            last_line_cos = cos_df.at[cos_df_len - 1, AdvisorConst.COSINE_SIMILARITY]
            if last_line_cos < AdvisorConst.ACCURACY_THRESHOLD:
                return AdvisorResult(True, AdvisorConst.CONSISTENCY_DETECTION,
                                     str(index), AdvisorConst.PROBLEM_SUGGEST)
            return AdvisorResult(True, AdvisorConst.CONSISTENCY_DETECTION,
                                 str(index), AdvisorConst.DEVIATION_SUGGEST)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import datetime
import os
import re
from multiprocessing import Process

import pytz
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from msprobe.core.common.const import MonitorConst
from msprobe.core.common.file_utils import read_csv, create_directory, remove_path, recursive_chmod
from msprobe.core.common.utils import check_process_num
from msprobe.core.common.decorator import recursion_depth_decorator
from msprobe.core.monitor.utils import get_target_output_dir
from msprobe.pytorch.common.log import logger


all_data_type_list = [
    "actv", "actv_grad", "exp_avg", "exp_avg_sq",
    "grad_unreduced", "grad_reduced", "param_origin", "param_updated"
]
CSV_FILE_SUFFIX = r"_\d+-\d+\.csv"


def parse_step_line(line, ops):
    vp_id = line["vpp_stage"]
    module_name = line[MonitorConst.HEADER_NAME]
    step = line["step"]
    vpp_name = f"vp{vp_id}:{module_name}"
    if 'micro_step' in line:
        vpp_name = f'{vpp_name}{MonitorConst.NAME_SEP}micro{line["micro_step"]}'
    ops_result = {}
    for op in ops:
        ops_result[op] = line[op]
    return vpp_name, step, ops_result


def parse_step_fn(filepath):
    data = read_csv(filepath)
    ops = [k for k in data.keys() if k in MonitorConst.OP_LIST[:-2]]
    parse_step_result = {}

    for _, line in data.iterrows():
        vpp_name, step, ops_result = parse_step_line(line, ops)
        if vpp_name not in parse_step_result:
            parse_step_result[vpp_name] = {}
        if step in parse_step_result[vpp_name]:
            raise Exception(f"duplicated step({step})")
        parse_step_result[vpp_name][step] = ops_result
    return parse_step_result


def write_step(output_dirpath, parse_step_result, rank, data_type):
    tb_output_path = os.path.join(output_dirpath, f"rank{rank}", data_type)
    if os.path.exists(tb_output_path):
        remove_path(tb_output_path)
        logger.warning(f"existing path {tb_output_path} will be recovered")
    writer = SummaryWriter(tb_output_path)
    for vpp_name, step_data_dict in parse_step_result.items():
        step_data_list = [(step, ops) for step, ops in step_data_dict.items()]
        step_data_list.sort(key=lambda x: x[0])
        for step_data in step_data_list:
            step = step_data[0]
            ops = step_data[1]
            for op, value in ops.items():
                tag = f"{vpp_name}/{op}"
                writer.add_scalar(tag, value, step)
    writer.close()


@recursion_depth_decorator("update_dict", max_depth=50)
def update_dict(dict1, dict2):
    for key, value in dict2.items():
        if key in dict1:
            if isinstance(dict1[key], dict) and isinstance(value, dict):
                try:
                    update_dict(dict1[key], value)
                except Exception as e:
                    raise Exception(f"Error updating nested dict failed at key '{key}': {e}") from e
            else:
                raise Exception(f"duplicate key: {key}")
        else:
            dict1[key] = value
    return dict1


def csv2tb_by_step_work(target_output_dirs, output_dirpath, data_type_list):
    for directory in tqdm(target_output_dirs):
        dirpath = directory["path"]
        rank = directory["rank"]
        for data_type in data_type_list:
            all_step_result = {}
            for filename in os.listdir(dirpath):
                if not re.match(f"{data_type}{CSV_FILE_SUFFIX}", filename):
                    continue
                filepath = os.path.join(dirpath, filename)
                try:
                    parse_step_result = parse_step_fn(filepath)
                except Exception as e:
                    logger.error(f"csv2tensorboard parse {filepath} failed \n {e}")
                    break

                all_step_result = update_dict(all_step_result, parse_step_result)
            if all_step_result:
                write_step(output_dirpath, all_step_result, rank, data_type)


def check_data_type_list(data_type_list):
    if data_type_list is None:
        logger.info(f"data_type_list is None, use default all_data_type_list: {all_data_type_list}")
        return
    if not isinstance(data_type_list, list):
        raise ValueError(f"data_type_list({data_type_list}) is not a list")
    for data_type in data_type_list:
        if data_type not in all_data_type_list:
            raise ValueError(f"data type({data_type}) is not supported, supported data type: {all_data_type_list}")


def csv2tensorboard_by_step(
        monitor_path,
        time_start=None,
        time_end=None,
        process_num=1,
        data_type_list=None,
        output_dirpath=None
):
    check_process_num(process_num)
    check_data_type_list(data_type_list)
    target_output_dirs = get_target_output_dir(monitor_path, time_start, time_end)
    target_output_dirs = [{"rank": rank, "path": path} for rank, path in target_output_dirs.items()]
    if output_dirpath is None:
        local_tz = pytz.timezone("Asia/Shanghai")  # 根据需要调整为目标时区
        cur_time = datetime.datetime.now(local_tz).strftime("%b%d_%H-%M-%S")
        output_dirpath = os.path.join(monitor_path, f"{cur_time}-csv2tensorboard_by_step")
    create_directory(output_dirpath)

    task_num = len(target_output_dirs)
    task_num_per_pro = task_num // process_num
    target_data_type = data_type_list if data_type_list else all_data_type_list

    processes = []
    for pro_id in range(process_num):
        task_start_id = pro_id * task_num_per_pro
        task_end_id = (pro_id + 1) * task_num_per_pro if pro_id != process_num - 1 else task_num
        task_dirs = target_output_dirs[task_start_id: task_end_id]

        p = Process(target=csv2tb_by_step_work, args=(task_dirs, output_dirpath, target_data_type))
        processes.append(p)
        p.start()
    for p in processes:
        p.join()
    recursive_chmod(output_dirpath)
    logger.info(f"output has been saved to: {output_dirpath}")



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from enum import Enum
import re
from msprobe.visualization.builder.msprobe_adapter import op_patterns
from msprobe.core.common.log import logger


class NodeOp(Enum):
    module = 0
    function_api = 1
    api_collection = 9

    @staticmethod
    def get_node_op(node_name: str):
        """
        基于代表节点的字符串，解析节点种类
        """
        for op in NodeOp:
            index = op.value
            if index < 0 or index >= len(op_patterns):
                continue
            pattern = op_patterns[index]
            if re.match(pattern, node_name):
                return op
        logger.warning(f"Cannot parse node_name {node_name} into NodeOp, default parsing as module.")
        return NodeOp.module



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.pytorch.api_accuracy_checker.compare.algorithm import compare_bool_tensor
from msprobe.pytorch.api_accuracy_checker.precision_standard.base_standard import BaseCompare


class BinaryCompare(BaseCompare):
    """
    Binary comparison class for comparing boolean tensors.

    This class is designed to compare the output of a binary operation between a benchmark and a device.
    It calculates the error rate of the comparison and provides a simple metric for assessing the accuracy.

    Attributes:
        bench_output (np.ndarray): The output from the benchmark.
        device_output (np.ndarray): The output from the device.
        compare_column (object): The column object to store comparison results.
        dtype (torch.dtype): The data type of the outputs.

    Methods:
        _compute_metrics(): Computes the comparison metrics, specifically the error rate.

    Note:
        This class assumes that the input data is an instance of InputData containing the benchmark output,
        device output, comparison column, and data type. The outputs are expected to be boolean tensors.

    See Also:
        BaseCompare: The base class for comparison classes.
        compare_bool_tensor: The function used to compare boolean tensors.
    """
    def __init__(self, input_data):
        super(BinaryCompare, self).__init__(input_data)

    def _pre_compare(self):
        pass

    def _compute_metrics(self):
        """
        Computes the error rate metric for the comparison between benchmark and device outputs.

        This method calculates the proportion of mismatches between the benchmark output and the device output.
        It uses the `compare_bool_tensor` function to compare the two tensors and extract the error rate.

        Returns:
        dict: A dictionary containing the computed error rate metric.
            The dictionary has the following key:
            - "error_rate": The proportion of mismatches between the benchmark and device outputs.
        """
        error_rate, _, _ = compare_bool_tensor(self.bench_output, self.device_output)

        return {
            "error_rate": error_rate
        }



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re

import torch

from msprobe.pytorch.monitor.features import get_max, get_min, get_zeros, get_nans, get_norm, get_mean
from msprobe.pytorch.monitor.features import cal_entropy, cal_stable_rank
from msprobe.pytorch.monitor.utils import get_nan_tensor


def get_summary_writer_tag_name(module_or_param_name: str, tag: str, rank):
    if rank is None:
        return f"{module_or_param_name}/{tag}"
    else:
        return f"{module_or_param_name}/rank{rank}/{tag}"


def squash_param_name(param_name, enable=True):
    if not enable:
        return param_name
    name = ''
    for pattern in ['^.*\.(layers?\..*)', '^.*\.(embeddings?\..*)', '^.*\.(final.*)', '^.*\.(output.*)',
                    '^.*\.(norm.*)']:
        match = re.findall(pattern, param_name)
        if match:
            name += match[0]
            break
    if name == '':
        name = param_name
    return name


# 用于存储所有metric实现类的注册表
config_metric_registry = {}


def register_config_metric(key, cls=None):
    """装饰器 用于注册Metric的实现类"""
    if cls is None:
        # 无参数时，返回装饰器函数
        return lambda cls_: register_config_metric(key, cls_)
    config_metric_registry[key] = cls()
    return cls


class TensorMetrics:
    fun_map = {"norm": get_norm, "max": get_max, "min": get_min, "mean": get_mean}

    def __init__(self) -> None:
        self.metrics = {}  # tensor_tag --> []
        self.cur_idx = {}

    def stat_insert(self, tensor, stat_ops, module_name, tensor_name, rank):
        """get stats and insert into metrics dictionary"""
        prefix = get_summary_writer_tag_name(module_name, tensor_name, rank)
        for stat_op in stat_ops:
            y = TensorMetrics.fun_map[stat_op](tensor)
            key = f"{prefix}_{stat_op}"
            if key not in self.metrics:
                self.metrics[key] = []
                self.cur_idx[key] = 0
            self.metrics[key].append(y)

    def flush(self, tb_writer):
        for key, metric_list in self.metrics.items():
            start = self.cur_idx[key]
            for v in metric_list[start:]:
                tb_writer.add_scalar(key, v.item(), global_step=self.cur_idx[key])
                self.cur_idx[key] += 1


class Metric(object):
    @staticmethod
    def get_metric_value(tensor, eps):
        NotImplementedError

    def get_metric(self, tensor, eps):
        try:
            return self.get_metric_value(tensor, eps)
        except RuntimeError as e:
            return torch.tensor(torch.nan).to(tensor.device)


@register_config_metric("min")
class MinMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_min(tensor)


@register_config_metric("mean")
class MeanMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_mean(tensor)


@register_config_metric("max")
class MaxMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_max(tensor)


@register_config_metric("norm")
class NormMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_norm(tensor)


@register_config_metric("zeros")
class ZerosMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_zeros(tensor, eps)


@register_config_metric("nans")
class NaNsMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return get_nans(tensor)


@register_config_metric("id")
class IdentMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        if tensor.dim() != 0:
            return None
        return tensor


@register_config_metric("shape")
class ShapeMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return tensor.shape


@register_config_metric("dtype")
class DtypeMetric(Metric):
    @staticmethod
    def get_metric_value(tensor, eps):
        return tensor.dtype


def get_metrics(ops, tag2tensor, eps, out_dict=None):
    """
    :param ops: ["op1", "op2"]
    :param tag2tensor: {
    '0:fc.input:0/actv': torch.randn([3, 4]),
    '0:fc.output:0/actv': torch.randn([3, 3])
    }
    :param eps: float 1e-8
    :param out_dict:{
    '0:fc.input:0/actv': {"op1": op1(torch.randn([3, 4])), "op2": op2(torch.randn([3, 4]))}
    '0:fc.output:0/actv': {"op1": op1(torch.randn([3, 3])), "op2": op2(torch.randn([3, 3]))}
    }
    :return: out_dict
    """
    if out_dict is None:
        out_dict = {}
    for tag, tensor in tag2tensor.items():
        if tag not in out_dict:
            out_dict[tag] = {}
        if not torch.is_tensor(tensor):
            # Non-tensor in/output filled with nan.
            out_dict[tag].update({metric_name: get_nan_tensor() for metric_name in ops})
            continue
        for metric_name in ops:
            fun_metric = config_metric_registry.get(metric_name)
            out_dict[tag][metric_name] = fun_metric.get_metric(tensor, eps)
    return out_dict


def get_sr_metric(tag2tensor, out_dict=None):
    if out_dict is None:
        out_dict = {}
    for tag, tensor in tag2tensor.items():
        if "sr" not in tag:
            continue
        if tag not in out_dict:
            out_dict[tag] = {}
        sr, eig = cal_stable_rank(tensor)
        out_dict[tag]['sr'] = sr
        out_dict[tag]['kernel_norm'] = eig


def get_entropy_metric(tag2tensor, out_dict=None):
    if out_dict is None:
        out_dict = {}
    for tag, tensor in tag2tensor.items():
        if tag not in out_dict:
            out_dict[tag] = {}
        entropy, softmax_max = cal_entropy(tensor)
        out_dict[tag]['entropy'] = entropy
        out_dict[tag]['softmax_max'] = softmax_max



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import time
from collections import defaultdict
import os
from itertools import dropwhile, chain

from msprobe.core.common.file_utils import save_json, create_directory
from msprobe.core.common.log import logger
from msprobe.core.common.const import Const
from msprobe.core.compare.utils import compare_distributed_inner
from msprobe.core.compare.find_first.utils import (RankPath, FileCache, is_communication_op, is_ignore_op,
                                                   DiffAnalyseConst, analyze_diff_in_group)
from msprobe.core.compare.find_first.graph import DataNode, CommunicationNode


class DiffAnalyzer:
    def __init__(self, npu_path, bench_path, output_path):
        self._bench_path = bench_path
        self._npu_path = npu_path
        self._output_path = output_path
        self._paths = {}
        self._diff_nodes = []  # 记录所有异常节点
        self._cache = FileCache()
        self._first_comm_nodes = {}  # 记录各rank下首个通信节点的node_id
        self._after_comm_diffs = {}  # 记录各rank下发生在通信节点之后的异常计算节点
        self._rank_comm_nodes_dict = {}  # 记录各rank的通信节点

    def analyze(self):
        self._pre_process()
        for analyze_func in [self._pre_analyze, self._analyze, self._post_analyze]:
            analyze_func()
            if self._diff_nodes:
                self._gen_analyze_info()
                return
        logger.info('Cannot find any diff node, no need to generate analyze file.')

    def _pre_process(self):
        logger.info("Start comparing data ......")
        compare_distributed_inner(self._npu_path, self._bench_path, self._output_path, first_diff_analyze=True)
        self._resolve_input_path(self._output_path)
        logger.info("Pre Process completed.")

    """
    这里需要生成stack，但是直接用dict中自带就行，在op_items.NPU_Stack_Info中
    """
    def _resolve_input_path(self, result_input_path):
        contents = os.listdir(result_input_path)
        rank_paths = {}
        
        for path in contents:
            # 检查文件名是否符合compare_result_rank{rank_id}_{timestamp}.json格式
            if not path.startswith('compare_result_rank'):
                continue
            if not path.endswith('.json'):
                continue
                
            # 从文件名中提取rank_id
            try:
                path_ele_list = path.split('_')
                if len(path_ele_list) <= 2:
                    continue
                rank_part = path_ele_list[2]
                if not rank_part.startswith('rank'):
                    continue
                rank_str = rank_part.strip('rank')  # 去掉'rank'前缀
                rank = int(rank_str) if rank_str else 0
            except (IndexError, ValueError):
                continue
                
            # 构建完整的json文件路径
            dump_path = os.path.join(result_input_path, path)
            rank_paths[rank] = RankPath(rank, dump_path)
            
        # 按照rank id排序后添加到self._paths中
        for rank in sorted(rank_paths.keys()):
            self._paths[rank] = rank_paths[rank]

    def _pre_analyze(self):
        logger.info('Start searching diff node before communication.')
        for path in self._paths.values():
            dump_data = self._cache.load_json(path.dump_path)
            if not dump_data:
                logger.warning(f'Rank {path.rank} has no dump data!')
                continue
            for op_name, op_data in dump_data.items():
                if is_ignore_op(op_name):
                    continue
                if is_communication_op(op_name):
                    self._first_comm_nodes[path.rank] = op_name
                    break
                data_node = DataNode(op_name, path.rank, op_data)
                if data_node.is_diff:
                    self._diff_nodes.append(data_node)
                    break

    def _analyze(self):
        logger.info('Start searching diff node during communication.')
        self._rank_comm_nodes_dict = {rank: self._analyze_comm_nodes(rank) for rank in self._paths}
        self._connect_comm_nodes()
        self._pruning()
        self._search_first_diff()

    def _post_analyze(self):
        logger.info('Start searching diff node after communication.')
        for nodes in self._after_comm_diffs.values():
            if nodes:
                self._diff_nodes.append(nodes[0])

    def _connect_comm_nodes(self):
        searched_ranks = set()
        for rank, nodes in list(self._rank_comm_nodes_dict.items())[:-1]:
            searched_ranks.add(rank)
            seen_nodes = set()
            last_node = None
            for cur_node in nodes.values():
                is_overflow = last_node and hasattr(last_node, 'layer') and hasattr(cur_node, 'layer') and \
                last_node.layer >= cur_node.layer
                if is_overflow:
                    cur_node.layer = last_node.layer + 1
                conn_info = cur_node.find_connected_nodes()
                if not conn_info.get('ranks'):
                    conn_info['ranks'] = self._rank_comm_nodes_dict.keys()
                last_node = cur_node
                if not self._find_connection(conn_info, cur_node, searched_ranks, seen_nodes):
                    logger.debug(f'Cannot find connected communication node for "{cur_node.node_id}".')

    def _find_connection(self, conn_info, cur_node, searched_ranks, seen_nodes):
        def connect(search_node):
            seen_nodes.add(search_node.node_id)
            if search_node.type == DiffAnalyseConst.DST:
                cur_node.add_dst(search_node)
            elif search_node.type == DiffAnalyseConst.SRC:
                search_node.layer = cur_node.layer
                search_node.add_dst(cur_node)
            else:
                cur_node.add_link(search_node)

        found = cur_node.connected
        for connected_rank in conn_info['ranks']:
            if connected_rank in searched_ranks:
                continue
            tar_id_prefix = f'{connected_rank}.{conn_info["api"]}'
            for search_id, search_node in self._rank_comm_nodes_dict[connected_rank].items():
                if search_id in seen_nodes:
                    continue
                if not (search_id.startswith(tar_id_prefix) and search_node.type == conn_info.get('type')):
                    continue
                search_conn_ranks = search_node.find_connected_nodes().get('ranks')
                if ((not search_conn_ranks and search_node.api not in DiffAnalyseConst.DIRECTED_API) or
                    cur_node.rank in search_conn_ranks):  # 有些无向通信算子没有填ProcessGroup，默认连接所有rank
                    connect(search_node)
                    found = True
                    break
        return found

    def _analyze_comm_nodes(self, rank):
        path = self._paths[rank]
        data = self._cache.load_json(path.dump_path)
        communication_nodes = {}
        if rank not in self._first_comm_nodes:  # 此rank没有通信节点
            return communication_nodes
        last_node_id = None  # 记录上一个通信节点的node_id
        compute_ops = []  # 记录两个通信节点之间的计算节点
        sub_layer = 0  # 记录两个通信算子之间异常计算节点的调用序数
        for op_name in dropwhile(lambda k: k != self._first_comm_nodes[rank], data):
            node_id = f'{rank}.{op_name}'
            op_data = data[op_name]
            if is_communication_op(op_name):
                comm_node = CommunicationNode(node_id, rank, DataNode(op_name, rank, op_data, sub_layer=sub_layer),
                                              compute_ops=compute_ops)
                if last_node_id:
                    communication_nodes.get(last_node_id).add_next(comm_node)
                communication_nodes[node_id] = comm_node
                last_node_id = node_id
                compute_ops = []
                sub_layer = 0
            elif not is_ignore_op(op_name):
                data_node = DataNode(op_name, rank, op_data, sub_layer=sub_layer)
                if data_node.is_diff:
                    compute_ops.append(data_node)
                sub_layer += 1
        if compute_ops:
            self._after_comm_diffs[rank] = compute_ops
        return communication_nodes

    def _pruning(self):
        deleted_node_id = []
        for nodes in self._rank_comm_nodes_dict.values():
            for node_id in list(nodes.keys()):
                node = nodes[node_id]
                if node.is_diff or node.compute_ops:
                    continue
                deleted_node_id.append(node_id)
                node.delete()
                del nodes[node_id]
        logger.debug(f'After pruning, following nodes are removed: [{", ".join(deleted_node_id)}]')

    def _search_first_diff(self):
        nodes_queues = []
        for comm_nodes in self._rank_comm_nodes_dict.values():
            nodes_queues.append(sorted(list(comm_nodes.values()), key=lambda x: x.layer))
        seen_nodes = set()

        def get_next_node(node_list):
            while node_list:
                next_node = node_list.pop(0)
                if next_node.node_id not in seen_nodes:
                    return next_node
            return None

        def find_all_members(ori_node):
            ids = get_relative_ids(ori_node)
            id_queue = list(chain(*[get_relative_ids(self._get_node_by_id(n_id)).difference(ids) for n_id in ids]))
            while id_queue:
                new_id = id_queue.pop(0)
                ids.add(new_id)
                id_queue.extend(get_relative_ids(self._get_node_by_id(new_id)).difference(ids))
            return ids

        def get_relative_ids(ori_node):
            if not ori_node:
                return set()
            return ({ori_node.node_id} | ori_node.link_nodes.keys() | ori_node.src_nodes.keys() |
                    ori_node.dst_nodes.keys())

        while any(nodes_queues):
            groups = []
            all_ids_in_groups = set()
            for nodes in nodes_queues:
                node = get_next_node(nodes)
                if not node:
                    continue
                if not groups or node.node_id not in all_ids_in_groups:
                    new_group = find_all_members(node)
                    groups.append(new_group)
                    all_ids_in_groups.update(new_group)
            for group in groups:
                seen_nodes.update(group)
                self._diff_nodes.extend(analyze_diff_in_group([self._get_node_by_id(n_id) for n_id in group]))
            if self._diff_nodes:
                # 找出所有layer和sub_layer最小的节点
                min_layer_sublayer = min((x.layer, x.sub_layer) for x in self._diff_nodes)
                self._diff_nodes = [
                                        node
                                        for node in self._diff_nodes
                                        if (node.layer, node.sub_layer) == min_layer_sublayer
                                   ]
                return

    def _get_node_by_id(self, node_id):
        splits = node_id.split(Const.SEP, 1)
        if len(splits) < 2 or not splits[0].isdigit():
            logger.error(f'invalid node_id {node_id}')
            raise RuntimeError(f'invalid node_id {node_id}')
        rank = int(splits[0])
        return self._rank_comm_nodes_dict.get(rank, {}).get(node_id)

    def _gen_analyze_info(self):
        if not os.path.exists(self._output_path):
            create_directory(self._output_path)
        file_name = f'diff_analyze_{time.time_ns()}.json'
        result_file = os.path.join(self._output_path, file_name)
        result_content = defaultdict(list)
        for node in self._diff_nodes:
            result_content[f'rank_{node.rank}'].append(node.gen_node_info(self._paths[node.rank]))
        save_json(result_file, result_content, 2)
        logger.info(f"The analyze result is saved in: {result_file}")



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the common function.
"""

from msprobe.msaccucmp.cmp_utils import common
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager, DD
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.path_check import check_name_valid
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_rule_parser import FusionRuleParser
from msprobe.msaccucmp.vector_cmp.fusion_manager import fusion_rule_parser


class TensorId:
    """
    The class for tensor id
    """

    def __init__(self: any, op_name: str, tensor_type: str, index: str) -> None:
        self.op_name = op_name
        self.tensor_type = tensor_type
        if not RegManager.match_pattern(RegManager.NUMBER_PATTERN, index):
            log.print_only_support_error('detail index', index, 'natural number')
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)
        self.index = int(index)

    def check_arguments_valid(self: any) -> None:
        """
        check arguments valid, if invalid, throw exception
        """
        ret = check_name_valid(self.op_name)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        if self.tensor_type not in ConstManager.SUPPORT_DETAIL_TYPE:
            log.print_only_support_error('detail type', self.tensor_type,
                                         ConstManager.SUPPORT_DETAIL_TYPE)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    def get_tensor_id(self: any) -> str:
        """
        Get tensor id
        """
        return "%s:%s:%d" % (self.op_name, self.tensor_type, self.index)

    def is_input(self: any) -> bool:
        """
        is input
        :return: true if input
        """
        return self.tensor_type == ConstManager.INPUT

    def get_file_prefix(self: any) -> str:
        """
        Get detail file name prefix
        :return str
        """
        return "%s_%s_%d" % (self.op_name.replace('/', '_'), self.tensor_type, self.index)

    def get_tensor_type_index(self: any) -> str:
        """
        Get detail file name prefix, like _input_0
        :return str
        """
        return "_%s_%d" % (self.tensor_type, self.index)


class DetailInfo:
    """
    The class for detail info
    """

    def __init__(self: any, tensor_id: TensorId, top_n: int, ignore_result: bool, max_line: int) -> None:
        self.tensor_id = tensor_id
        self.my_output_ops = ''
        self.ground_truth_ops = ''
        self.detail_format = ''
        self.top_n = top_n
        self.ignore_result = ignore_result
        self.max_line = max_line

    def check_arguments_valid(self: any) -> None:
        """
        check arguments valid, if invalid, throw exception
        """
        self.tensor_id.check_arguments_valid()
        const_manager = ConstManager()
        if self.top_n < const_manager.min_top_n or self.top_n > const_manager.max_top_n:
            log.print_out_of_range_error('', 'top n', self.top_n,
                                         '[%d, %d]' % (const_manager.min_top_n, const_manager.max_top_n))
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

    def get_detail_op(self: any, fusion_rule: FusionRuleParser) -> (FusionOp, list):
        """
        Get detail op by fusion rule
        :param: fusion_rule: the fusion rule
        :return: the fusion op, the fusion op list
        """
        fusion_op_list, fusion_op = fusion_rule.get_fusion_op_list(self.tensor_id.op_name)
        # get the map for {original_op_names, op_list}
        right_to_left_map = fusion_rule_parser.make_right_to_left_multi_map(
            fusion_op_list)
        my_output_ops_str, ground_truth_ops_str = fusion_rule_parser.make_left_and_right_string(right_to_left_map)
        # if right ops is empty, mark '*' indicates that the left op is a
        # new operator, and there is no operator on the right that matches it
        if ground_truth_ops_str == "":
            ground_truth_ops_str = '*'
        self.my_output_ops = my_output_ops_str.replace(',', ' ')
        self.ground_truth_ops = ground_truth_ops_str.replace(',', ' ')
        return fusion_op, fusion_op_list

    def set_detail_format(self: any, shape_str: str, tensor_format: int, ground_truth_format: str) -> None:
        """
        Set detail format by shape
        :param shape_str: the shape string
        :param tensor_format: tensor format
        :param ground_truth_format: ground truth format
        """
        if common.contain_depth_dimension(tensor_format):
            self.detail_format = 'N C D H W'
        else:
            if tensor_format == DD.FORMAT_FRACTAL_Z:
                self.detail_format = ' '.join(ground_truth_format)
            else:
                self.detail_format = "N C H W" if shape_str != "()" else "ID"

    def check_and_set_format(self: any, shape_str: str, tensor_format: int, ground_truth_format: int) -> None:
        """
        Check and Set detail format by shape
        :param shape_str: the shape string
        :param tensor_format: tensor format
        :param ground_truth_format: ground truth format
        """
        if tensor_format != ground_truth_format:
            log.print_error_log("NPUDump tensor format not match Ground truth tensor format!"
                                "Cannot be directly compared.")
            raise CompareError(CompareError.MSACCUCMP_INVALID_FORMAT_ERROR)
        if common.contain_depth_dimension(tensor_format):
            self.detail_format = 'N C D H W'
        else:
            self.detail_format = "N C H W" if shape_str != "()" else "ID"

    def set_detail_ops(self: any, my_out_ops: str, ground_truth_ops: str) -> None:
        """
        Check and Set detail format by shape
        :param my_out_ops: the opname of my out tensor
        :param ground_truth_ops: the opname of ground truth tensor
        """
        self.my_output_ops = my_out_ops.replace(',', ' ')
        self.ground_truth_ops = ground_truth_ops.replace(',', ' ')

    def make_detail_header(self: any) -> str:
        """
        Make detail header
        """
        return "Index,%s,NPUDump,GroundTruth,AbsoluteError,RelativeError\n" % self.detail_format

    def get_detail_info(self: any) -> str:
        """
        Get detail_info
        """
        return "NPUDump:%s\nGroundTruth:%s\nFormat:%s\n" % (self.my_output_ops, self.ground_truth_ops,
                                                            self.detail_format)



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file is used to parse the HDF5 file format.
"""
import re
from enum import Enum
import collections

from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class DataSetType(Enum):
    """
    The enum for pytorch dump data type
    """
    TENSOR = 0
    VEC_TENSOR = 1
    VEC_I64 = 2
    I64 = 3
    BOOL = 4
    DOUBLE = 5
    SCALER = 6
    TYPE_AND_SIZED = 7
    OPT_INT64 = 8
    OPT_SCALER = 9
    VEC_VEC_I64 = 10
    GEOMETRY = 11
    SIZE = 12
    SCALER_TYPE = 13


def _open_h5py_file(file_path: str, model: str = 'r') -> any:
    """
    Open the HDF5 file
    """
    import h5py

    return h5py.File(file_path, model)


class Hdf5Parser:
    """
    The class for HDF5 file parse
    """
    MAX_OP_NUM = 1000000
    DATASET_PATH_PATTERN = r"(/[a-zA-Z0-9_]*)(/[0-9]*)(/[a-z]*)"
    INPUT = "/input"
    GOLDEN_DUMP_FILE = "GOLDEN_DUMP"
    MY_DUMP_FILE = "MY_DUMP"

    def __init__(self: any, file_path: str, dump_file_type: str, mapping_list: list) -> None:
        self.file_handle = None
        self.dump_file_type = dump_file_type
        self.file_path = file_path
        self.order_ext_opname_map = collections.defaultdict(list)
        self.ext_opname_dataset_map = collections.defaultdict(list)
        self.device_type = None
        self.need_compare_input = False
        self._mapping_list = mapping_list

    @staticmethod
    def _is_multimap(op_name: str, mapping_set: list) -> bool:
        if op_name in mapping_set:
            return True
        return False

    def open_file(self: any, model: str) -> int:
        """
        Open file and check exception
        """
        try:
            self.file_handle = _open_h5py_file(self.file_path, model)
            return CompareError.MSACCUCMP_NONE_ERROR
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError):
            self.file_handle = None
            log.print_error_log("Open {} failed!".format(self.file_path))
            return CompareError.MSACCUCMP_OPEN_FILE_ERROR
        finally:
            pass

    def close_file(self: any) -> None:
        """
        Close the file by file_handle
        """
        if self.file_handle is not None:
            try:
                self.file_handle.close()
            except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                    MemoryError):
                log.print_error_log("Close {} failed!".format(self.file_path))
                return
            self.file_handle = None

    def get_dump_data_attr(self: any, dataset_path: str, attr_type: str) -> any:
        """
        Get the dump data attribute by dataset_path
        :dataset_path: the path of dataset
        :attr_type:
        """
        if self.file_handle is None:
            log.print_error_log("the handle of {} is invalid!".format(self.file_path))
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)
        try:
            attrs = self.file_handle[dataset_path].attrs
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as err:
            log.print_error_log("Read dataset attr:{} failed!"
                                .format(dataset_path))
            raise CompareError(CompareError.MSACCUCMP_PARSE_DUMP_FILE_ERROR) from err
        if attrs:
            return True, attrs.get(attr_type)
        return False, ''

    def get_dump_data(self: any, dataset_path: str) -> any:
        """
        Get the dump data by dataset_path
        :dataset_path: the path of dataset
        """
        if self.file_handle is None:
            log.print_error_log("the handle of {} is invalid!".format(self.file_path))
            raise CompareError(CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR)
        try:
            if self._dataset_path_valid(dataset_path):
                return self.file_handle[dataset_path][()]
            return []
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as err:
            log.print_error_log("Read dataset:{} failed!"
                                .format(dataset_path))
            raise CompareError(CompareError.MSACCUCMP_PARSE_DUMP_FILE_ERROR) from err
        finally:
            pass

    def get_all_orders(self: any) -> list:
        """
        Get all orders string used to compare dump data.
        """
        return list(self.order_ext_opname_map.keys())

    def get_order_by_ext_opname(self: any, ext_opname: str) -> int:
        """
        Get the order corresponding to the ext_opname.
        """
        for order, ext_opname_group in self.order_ext_opname_map.items():
            if ext_opname in ext_opname_group:
                return order
        # Unmatched items. To display them at the end of the result file after sorting,
        # order must be the maximum value plus 1.
        if self.get_all_orders():
            return max(self.get_all_orders()) + 1
        return 0

    def get_ext_opname_group_by_order(self: any, order: int) -> list:
        """
        Get the list of op names corresponding to the order.
        """
        return self.order_ext_opname_map.get(order, [])

    def parse_dump_file(self: any) -> int:
        """
        Entry for parsing dump files.
        """
        ret = self.open_file('r')
        if ret == CompareError.MSACCUCMP_NONE_ERROR:
            self._generate_order_ext_opname_map()
            self._parse_all_dataset()
        return ret

    def have_dataset(self: any, ext_opname: str, dataset_path: str) -> bool:
        """
        Check whether the dataset path is included.
        :ext_opname: extended op name
        :dataset_path: dataset path in dump file
        """
        if dataset_path in self.ext_opname_dataset_map.get(ext_opname, []):
            return True
        if ConstManager.DELIMITER not in ext_opname:
            log.print_error_log(f": not in {ext_opname}, please check.")
            raise CompareError(CompareError.MSACCUCMP_NAME_ERROR)
        op_name, _ = ext_opname.split(ConstManager.DELIMITER, 1)
        # get 4 from '/AddmmBackward/4/input/mat1'

        _, _, order, _ = dataset_path.split('/', 3)
        all_ext_opname = self.get_ext_opname_group_by_order(int(order))
        for item in all_ext_opname:
            if item.startswith("{}:".format(op_name)) and \
                    dataset_path in self.ext_opname_dataset_map.get(item, []):
                return True
        return False

    def file_is_empty(self: any) -> bool:
        """
        Check whether the file is empty.
        """
        return self.file_handle is None \
               or self.file_handle.keys()

    def is_load_mode(self: any) -> bool:
        """
        Check whether the load comparison mode.
        """
        return not self.need_compare_input

    def _check_value(self: any, op_order_list: list) -> None:
        if len(op_order_list) > self.MAX_OP_NUM:
            log.print_error_log("The number of ops in {} exceeds the range!"
                                .format(self.file_path))
            raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)

    def _get_mapping_set(self: any, op_name: str) -> list:
        for mapping_set in self._mapping_list:
            if op_name in mapping_set and len(mapping_set) > 1:
                return mapping_set
        return []

    def _is_parsed(self: any, op_name: str) -> bool:
        for ext_op_name_list in self.order_ext_opname_map.values():
            for ext_op_name in ext_op_name_list:
                if ext_op_name.startswith(op_name):
                    return True
        return False

    def _gen_ext_opname_map_special(self: any, op_name: str, multimap_set: list) -> dict:
        order_ext_opname_map = collections.defaultdict(list)
        if self._is_parsed(op_name):
            return order_ext_opname_map
        op_order_list = []
        for op in multimap_set:
            if op in self.file_handle.keys():
                op_order_list.extend(list(map(int, self.file_handle[op].keys())))
        op_order_list.sort()
        for index, op_order in enumerate(op_order_list):
            for op in multimap_set:
                if op in self.file_handle.keys() and str(op_order) in self.file_handle[op].keys():
                    order_ext_opname_map[op_order] = ["{}:{}".format(op, index)]
        return order_ext_opname_map

    def _gen_single_order_ext_opname_map(self: any, op_name: str) -> dict:
        """
        Get the map of order to extended op name.
        """
        order_ext_opname_map = collections.defaultdict(list)
        mapping_set = self._get_mapping_set(op_name)
        if self.file_handle is None:
            return order_ext_opname_map
        try:
            if self._is_multimap(op_name, mapping_set):
                return self._gen_ext_opname_map_special(op_name, mapping_set)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError):
            log.print_error_log("construct order_ext_map by {} failed:!"
                                .format(op_name))
            return order_ext_opname_map

        # sort by the order of the op executed multi times
        op_order_list = list(map(int, self.file_handle[op_name].keys()))
        op_order_list.sort()
        self._check_value(op_order_list)
        for index, op_order in enumerate(op_order_list):
            order_ext_opname_map[op_order] = ["{}:{}".format(op_name, index)]
        return order_ext_opname_map

    def _generate_order_ext_opname_map(self: any) -> None:
        """
        Generate the mapping between order and extended operator names in dump data.
        """
        if self.file_handle is None:
            return
        for op_name in self.file_handle.keys():
            order_op_map = self._gen_single_order_ext_opname_map(op_name)
            utils.merge_dict(self.order_ext_opname_map, order_op_map)

    def _dataset_path_valid(self: any, dataset_path: str) -> bool:
        match_result = re.search(self.DATASET_PATH_PATTERN, dataset_path)
        if match_result is not None:
            return True
        return False

    def _data_path_is_input(self: any, curr_group_path: str) -> bool:
        """
        Check whether the data path contains the input flag.
        :curr_group_path: the path of of group or dataset
        """
        match_result = re.search(self.DATASET_PATH_PATTERN, curr_group_path)
        if match_result is not None and match_result.group(3) == self.INPUT:
            return True
        return False

    def _parse_one_dataset(self: any, data_type: int, ext_opname: str, curr_group_path: str) -> None:
        # compare Tensor data only.
        if data_type != DataSetType.TENSOR.value:
            return
        self.ext_opname_dataset_map[ext_opname].append(curr_group_path)
        if self.device_type and self.device_type != utils_type.DeviceType.CPU.value:
            return
        attr_ok, device_type = self.get_dump_data_attr(curr_group_path,
                                                       utils_type.DatasetAttr.DeviceType.name)
        if attr_ok:
            self.device_type = device_type

    def _need_compare(self: any, curr_group_path: str) -> bool:
        if not self._data_path_is_input(curr_group_path):
            return True
        if self.dump_file_type == self.MY_DUMP_FILE:
            self.need_compare_input = True
            return True
        if self.dump_file_type == self.GOLDEN_DUMP_FILE and not self.need_compare_input:
            return False
        return True

    def _parse_dataset_recursively(self: any, ext_opname: str, curr_group_path: str) -> None:
        """
        Get the data_set path of one op recursively
        :ext_opname: example. Adamm_1, 1 is the execution order of the op Adamm in the model.
        :curr_group_path: current hierarchical path found recursively
        """
        if self.file_handle is None:
            self.ext_opname_dataset_map.clear()

        if not self._need_compare(curr_group_path):
            return
        # only dataset have the attribute 'Type'. The group dose not.
        have_attr, data_type = self.get_dump_data_attr(curr_group_path, utils_type.DatasetAttr.Type.name)
        if have_attr:
            self._parse_one_dataset(data_type, ext_opname, curr_group_path)
            return
        try:
            # find all the datasets recursively by path
            if self.file_handle[curr_group_path].keys():
                for next_group_name in self.file_handle[curr_group_path].keys():
                    next_group_path = "{}/{}".format(curr_group_path, next_group_name)
                    self._parse_dataset_recursively(ext_opname, next_group_path)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError):
            log.print_error_log("parse_dataset failed, group path is:{}!"
                                .format(curr_group_path))
        finally:
            pass

    def _parse_all_dataset(self: any) -> None:
        """
        Parse the hdf5 dump file.
        Obtain all dataset paths from this file.
        """
        if self.file_handle is None:
            return

        for order in self.order_ext_opname_map.keys():
            ext_opname_group = self.order_ext_opname_map.get(order, [])
            for ext_opname in ext_opname_group:
                if ConstManager.DELIMITER not in ext_opname:
                    log.print_error_log(f": not in {ext_opname}, please check.")
                    raise CompareError(CompareError.MSACCUCMP_NAME_ERROR)
                op_name, _ = ext_opname.split(ConstManager.DELIMITER, 1)
                group_path = "/{}/{}".format(op_name, order)
                self._parse_dataset_recursively(ext_opname, group_path)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.file_utils import create_directory, save_json
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.dump.debugger.debugger_config import DebuggerConfig


class KernelGraphExceptionDump:

    def __init__(self, config: DebuggerConfig):
        self.dump_json = dict()
        self.dump_json["common_dump_settings"] = dict()
        self.dump_json["common_dump_settings"]["dump_mode"] = 0
        self.dump_json["common_dump_settings"]["path"] = ""
        self.dump_json["common_dump_settings"]["net_name"] = "Net"
        self.dump_json["common_dump_settings"]["iteration"] = "all"
        self.dump_json["common_dump_settings"]["saved_data"] = "tensor"
        self.dump_json["common_dump_settings"]["input_output"] = 0
        self.dump_json["common_dump_settings"]["kernels"] = []
        self.dump_json["common_dump_settings"]["support_device"] = [0, 1, 2, 3, 4, 5, 6, 7]
        self.dump_json["common_dump_settings"]["op_debug_mode"] = 4
        self.dump_json["common_dump_settings"]["file_format"] = "npy"
        self.dump_json["e2e_dump_settings"] = dict()
        self.dump_json["e2e_dump_settings"]["enable"] = not config.async_dump
        self.dump_json["e2e_dump_settings"]["trans_flag"] = True

        if config.stat_cal_mode and config.device_stat_precision_mode:
            self.dump_json["e2e_dump_settings"]["stat_calc_mode"] = config.stat_cal_mode
            self.dump_json["e2e_dump_settings"]["device_stat_precision_mode"] = config.device_stat_precision_mode
        self.dump_json["common_dump_settings"]["path"] = config.dump_path
        if len(config.step) > 0:
            logger.warning("Step would change to all in this task.")
        if len(config.rank) > 0:
            self.dump_json["common_dump_settings"]["support_device"] = config.rank

    def handle(self):
        json_path = self.dump_json["common_dump_settings"]["path"]
        create_directory(json_path)
        json_path = os.path.join(json_path, "kernel_graph_exception_check.json")
        save_json(json_path, self.dump_json, indent=4)
        logger.info(json_path + " has been created.")
        os.environ["MINDSPORE_DUMP_CONFIG"] = json_path



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import math
import json
from msprobe.core.common.const import CompareConst, Const
from msprobe.visualization.utils import ToolTip, GraphConst, str2float


class ModeAdapter:
    def __init__(self, compare_mode, parallel_merge=False):
        self.compare_mode = compare_mode
        self.csv_data = []
        self.compare_nodes = []
        self.parallel_merge = parallel_merge

    @staticmethod
    def _is_invalid(value):
        if not isinstance(value, float):
            return False
        return math.isnan(value) or math.isinf(value)

    @staticmethod
    def _add_compare_data(node_data, compare_data_dict, key_list, headers):
        for key, data_info in node_data.items():
            if not isinstance(data_info, dict):
                continue
            compare_data = compare_data_dict.get(key)
            if compare_data:
                # 对应比对结果csv的列
                id_list = [headers.index(x) for x in key_list]
                ModeAdapter._match_data(data_info, compare_data, key_list, id_list)

    @staticmethod
    def _match_data(data_dict, compare_data, key_list, id_list):
        """
        绑定精度指标到node的input_data和output_data
        """
        if len(key_list) != len(id_list):
            return
        for id_val, key in zip(id_list, key_list):
            data_dict[key] = compare_data[id_val]

    @staticmethod
    def _check_list_len(data_list, len_num):
        if len(data_list) < len_num:
            raise ValueError(f"compare_data_dict_list must contain at least {len_num} items.")

    def parse_result(self, node, compare_data_dict, api_indicator):
        """
        根据结果返回precision_index
        """
        if self.compare_mode == GraphConst.MD5_COMPARE:
            key_list = GraphConst.MD5_INDEX_LIST
            headers = CompareConst.MD5_COMPARE_RESULT_HEADER
        elif self.compare_mode == GraphConst.SUMMARY_COMPARE:
            key_list = GraphConst.SUMMARY_INDEX_LIST
            headers = CompareConst.SUMMARY_COMPARE_RESULT_HEADER
        else:
            key_list = GraphConst.REAL_DATA_INDEX_LIST
            headers = CompareConst.COMPARE_RESULT_HEADER
        ModeAdapter._add_compare_data(node.input_data, compare_data_dict, key_list, headers)
        ModeAdapter._add_compare_data(node.output_data, compare_data_dict, key_list, headers)
        precision_index = GraphConst.COMPARE_INDICATOR_TO_PRECISION_INDEX_MAPPING.get(api_indicator,
                                                                                      CompareConst.PASS)
        precision_index = self._ignore_precision_index(node.id, precision_index)
        return precision_index

    def prepare_real_data(self, node):
        """
        为真实数据比较模式准备节点信息
        """
        if self.compare_mode == GraphConst.REAL_DATA_COMPARE:
            self.compare_nodes.append(node)
            return True
        return False

    def add_csv_data(self, compare_result_list):
        if self.compare_mode != GraphConst.REAL_DATA_COMPARE:
            return
        self.csv_data.extend(compare_result_list)

    def get_tool_tip(self):
        """
        用于前端展示字段的具体含义
        """
        if self.compare_mode == GraphConst.SUMMARY_COMPARE:
            tips = {
                CompareConst.MAX_DIFF: ToolTip.MAX_DIFF,
                CompareConst.MIN_DIFF: ToolTip.MIN_DIFF,
                CompareConst.MEAN_DIFF: ToolTip.MEAN_DIFF,
                CompareConst.NORM_DIFF: ToolTip.NORM_DIFF}
        elif self.compare_mode == GraphConst.MD5_COMPARE:
            tips = {Const.MD5: ToolTip.MD5}
        else:
            tips = {
                CompareConst.ONE_THOUSANDTH_ERR_RATIO: ToolTip.ONE_THOUSANDTH_ERR_RATIO,
                CompareConst.FIVE_THOUSANDTHS_ERR_RATIO: ToolTip.FIVE_THOUSANDTHS_ERR_RATIO,
                CompareConst.COSINE: ToolTip.COSINE,
                CompareConst.MAX_ABS_ERR: ToolTip.MAX_ABS_ERR,
                CompareConst.MAX_RELATIVE_ERR: ToolTip.MAX_RELATIVE_ERR}
        return json.dumps(tips)

    def _ignore_precision_index(self, node_id, precision_index):
        node_id_split = node_id.split(Const.SEP)
        if len(node_id_split) < 2:
            return precision_index
        if node_id.split(Const.SEP)[1] in GraphConst.IGNORE_PRECISION_INDEX:
            return GraphConst.MAX_INDEX_KEY if self.compare_mode == GraphConst.MD5_COMPARE else GraphConst.MIN_INDEX_KEY
        return precision_index



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from mindspore import dtype as mstype, Tensor

from msprobe.mindspore.monitor.features import FUNC_MAP, cal_entropy, cal_stable_rank


def get_single_metrics(op_list, tag, tensor, eps=1e-8, output=None):
    if output is None:
        output = {}
    if tag not in output:
        output[tag] = {}
    for op in op_list:
        func = FUNC_MAP.get(op)
        if op == "zeros":
            statistic = func(tensor, eps)
        else:
            statistic = func(tensor)
        if hasattr(statistic, "dtype") and statistic.dtype == mstype.bfloat16:
            statistic = float(statistic)
            statistic = Tensor(statistic)
        if isinstance(statistic, Tensor):
            output[tag][op] = statistic.astype(mstype.float32)
        else:
            output[tag][op] = statistic


def get_metrics(op_list, tag2tensor, eps, output=None):
    if output is None:
        output = {}
    for tag, tensor in tag2tensor.items():
        if tag not in output:
            output[tag] = {}
        get_single_metrics(op_list, tag, tensor, eps, output)
    return output


def get_summary_writer_tag_name(module_or_param_name: str, tag: str, rank):
    if rank is None:
        return f"{module_or_param_name}/{tag}"
    else:
        return f"{module_or_param_name}/rank{rank}/{tag}"


def step_accumulates_one(context, micro_batch_number):
    """
    :param context: ModuleHookContext
    :param micro_batch_number: mbs of training model.
    :return:
    """
    context.micro_step += 1
    if context.micro_step == micro_batch_number:
        context.micro_step = 0
        context.step += 1


def is_skip_step(step, start_step, step_interval, has_collect_times=0, collect_times=1e8):
    """
    If current step less than start_step or not reach step_interval, skip current step.
    :param step: current training step, int
    :param start_step: int
    :param step_interval: int
    :return: whether skip or not, bool
    """
    return step < start_step or (step - start_step) % step_interval != 0 or has_collect_times >= collect_times


def get_entropy_metric(tag2tensor, out_dict=None):
    if out_dict is None:
        out_dict = {}
    for tag, tensor in tag2tensor.items():
        if tag not in out_dict:
            out_dict[tag] = {}
        entropy, softmax = cal_entropy(tensor)
        out_dict[tag]["entropy"] = entropy
        out_dict[tag]["softmax"] = softmax
    return out_dict


def get_sr_metric(tag2tensor, out_dict=None):
    if out_dict is None:
        out_dict = {}
    for tag, tensor in tag2tensor.items():
        if "sr" not in tag:
            continue
        if tag not in out_dict:
            out_dict[tag] = {}
        sr, eig = cal_stable_rank(tensor)
        out_dict[tag]["sr"] = sr
        out_dict[tag]["eig"] = eig
    return out_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import CompareConst
from msprobe.pytorch.common.log import logger


class CompareColumn:
    __slots__ = [
        'bench_type', 'npu_type', 'shape', 'cosine_sim', 'max_abs_err', 'rel_err_hundredth',
        'rel_err_ten_thousandth', 'inf_nan_error_ratio', 'rel_err_ratio', 'abs_err_ratio', 
        'small_value_err_ratio', 'max_rel_error', 'mean_rel_error', 'rmse', 'eb', 'max_ulp_error', 
        'mean_ulp_error', 'ulp_error_proportion', 'error_rate', 'rel_err_thousandth'
    ]
    
    def __init__(self):
        self.bench_type = CompareConst.SPACE
        self.npu_type = CompareConst.SPACE
        self.shape = CompareConst.SPACE
        self.cosine_sim = CompareConst.SPACE
        self.max_abs_err = CompareConst.SPACE
        self.rel_err_hundredth = CompareConst.SPACE
        self.rel_err_thousandth = CompareConst.SPACE
        self.rel_err_ten_thousandth = CompareConst.SPACE
        self.error_rate = CompareConst.SPACE
        self.eb = CompareConst.SPACE
        self.rmse = CompareConst.SPACE
        self.small_value_err_ratio = CompareConst.SPACE
        self.max_rel_error = CompareConst.SPACE
        self.mean_rel_error = CompareConst.SPACE
        self.inf_nan_error_ratio = CompareConst.SPACE
        self.rel_err_ratio = CompareConst.SPACE
        self.abs_err_ratio = CompareConst.SPACE
        self.max_ulp_error = CompareConst.SPACE
        self.mean_ulp_error = CompareConst.SPACE
        self.ulp_error_proportion = CompareConst.SPACE

    def update(self, metrics):
        """
        Updates the object's attributes with the provided metrics.

        Args:
            metrics (dict): A dictionary containing attribute names and their corresponding values.

        Raises:
            AttributeError: If the metric key is not a valid attribute of CompareColumn.
        """
        for key, value in metrics.items():
            if value is None:
                continue
            if key not in self.__slots__:
                logger.error(f"The key '{key}' is not a valid attribute of CompareColumn.")
                continue
            setattr(self, key, value)

    def to_column_value(self, is_pass, message):
        return [self.bench_type, self.npu_type, self.shape, self.cosine_sim, self.max_abs_err, self.rel_err_hundredth,
                self.rel_err_thousandth, self.rel_err_ten_thousandth, self.error_rate, self.eb, self.rmse, 
                self.small_value_err_ratio, self.max_rel_error, self.mean_rel_error, self.inf_nan_error_ratio, 
                self.rel_err_ratio, self.abs_err_ratio, self.max_ulp_error, self.mean_ulp_error, 
                self.ulp_error_proportion, is_pass, message]


class ApiPrecisionOutputColumn:
    __slots__ = [
                'api_name', 'small_value_err_ratio', 'small_value_err_status', 'rmse_ratio', 'rmse_status', 
                'max_rel_err_ratio', 'max_rel_err_status', 'mean_rel_err_ratio', 'mean_rel_err_status', 'eb_ratio', 
                'eb_status', 'inf_nan_error_ratio', 'inf_nan_error_ratio_status', 'rel_err_ratio', 
                'rel_err_ratio_status', 'abs_err_ratio', 'abs_err_ratio_status', 'error_rate', 'error_rate_status', 
                'mean_ulp_err', 'ulp_err_proportion', 'ulp_err_proportion_ratio', 'ulp_err_status', 
                'rel_err_thousandth', 'rel_err_thousandth_status', 'compare_result', 'compare_algorithm', 
                'compare_message'
                ]
    
    def __init__(self):
        self.api_name = CompareConst.SPACE
        self.small_value_err_ratio = CompareConst.SPACE
        self.small_value_err_status = CompareConst.SPACE
        self.rmse_ratio = CompareConst.SPACE
        self.rmse_status = CompareConst.SPACE
        self.max_rel_err_ratio = CompareConst.SPACE
        self.max_rel_err_status = CompareConst.SPACE
        self.mean_rel_err_ratio = CompareConst.SPACE
        self.mean_rel_err_status = CompareConst.SPACE
        self.eb_ratio = CompareConst.SPACE
        self.eb_status = CompareConst.SPACE
        self.inf_nan_error_ratio = CompareConst.SPACE
        self.inf_nan_error_ratio_status = CompareConst.SPACE
        self.rel_err_ratio = CompareConst.SPACE
        self.rel_err_ratio_status = CompareConst.SPACE
        self.abs_err_ratio = CompareConst.SPACE
        self.abs_err_ratio_status = CompareConst.SPACE
        self.error_rate = CompareConst.SPACE
        self.error_rate_status = CompareConst.SPACE
        self.mean_ulp_err = CompareConst.SPACE
        self.ulp_err_proportion = CompareConst.SPACE
        self.ulp_err_proportion_ratio = CompareConst.SPACE
        self.ulp_err_status = CompareConst.SPACE
        self.rel_err_thousandth = CompareConst.SPACE
        self.rel_err_thousandth_status = CompareConst.SPACE
        self.compare_result = CompareConst.SPACE
        self.compare_algorithm = CompareConst.SPACE
        self.compare_message = CompareConst.SPACE

    def update(self, metrics):
        """
        Updates the object's attributes with the provided metrics.

        Args:
            metrics (dict): A dictionary containing attribute names and their corresponding values.

        Raises:
            AttributeError: If the metric key is not a valid attribute of CompareColumn.
        """
        for key, value in metrics.items():
            if value is None:
                continue
            if key not in self.__slots__:
                logger.error("The key '%s' is not a valid attribute of CompareColumn.", key)
                continue
            setattr(self, key, value)

    def to_column_value(self):
        return [self.api_name, self.small_value_err_ratio, self.small_value_err_status, self.rmse_ratio, 
                self.rmse_status, self.max_rel_err_ratio, self.max_rel_err_status, self.mean_rel_err_ratio, 
                self.mean_rel_err_status, self.eb_ratio, self.eb_status, self.inf_nan_error_ratio, 
                self.inf_nan_error_ratio_status, self.rel_err_ratio, self.rel_err_ratio_status, self.abs_err_ratio, 
                self.abs_err_ratio_status, self.error_rate, self.error_rate_status, self.mean_ulp_err, 
                self.ulp_err_proportion, self.ulp_err_proportion_ratio, self.ulp_err_status, self.rel_err_thousandth, 
                self.rel_err_thousandth_status, self.compare_result, self.compare_algorithm, self.compare_message]
        


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.pytorch.bench_functions.apply_adam_w import npu_apply_adam_w
from msprobe.pytorch.bench_functions.confusion_transpose import npu_confusion_transpose, \
    npu_confusion_transpose_backward
from msprobe.pytorch.bench_functions.fast_gelu import npu_fast_gelu, npu_fast_gelu_backward
from msprobe.pytorch.bench_functions.layer_norm_eval import npu_layer_norm_eval
from msprobe.pytorch.bench_functions.linear import npu_linear, npu_linear_backward
from msprobe.pytorch.bench_functions.matmul_backward import matmul_backward
from msprobe.pytorch.bench_functions.npu_fusion_attention import npu_fusion_attention, npu_fusion_attention_grad, \
    gpu_fusion_attention
from msprobe.pytorch.bench_functions.rms_norm import npu_rms_norm, npu_rms_norm_backward
from msprobe.pytorch.bench_functions.rotary_mul import npu_rotary_mul, npu_rotary_mul_backward
from msprobe.pytorch.bench_functions.scaled_mask_softmax import npu_scaled_masked_softmax, \
    npu_scaled_masked_softmax_backward
from msprobe.pytorch.bench_functions.swiglu import npu_swiglu, npu_swiglu_backward
from msprobe.pytorch.bench_functions.apply_adam import npu_apply_adam
from msprobe.pytorch.bench_functions.group_norm_silu import npu_group_norm_silu
from msprobe.pytorch.bench_functions.mish import npu_mish
from msprobe.pytorch.bench_functions.moe_gating_top_k_softmax import npu_moe_gating_top_k_softmax
from msprobe.pytorch.bench_functions.sort_v2 import npu_sort_v2
from msprobe.pytorch.common.utils import logger


class Register(dict):
    def __init__(self, *args, **kwargs):
        super(Register, self).__init__(*args, **kwargs)
        self._dict = {}

    def __call__(self, target_func_list):
        for target in target_func_list:
            self.register(target)
        return

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __str__(self):
        return str(self._dict)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()

    def register(self, target):

        def add_register_item(key, value):
            if key in self._dict:
                logger.warning(f"{value.__name__} has been registered before, so we will override it.")
            self[key] = value
            return value

        if callable(target):
            return add_register_item(target.__name__, target)
        else:
            raise Exception(f"The func {target} is not callable.")


# register for npu custom bench functions
npu_custom_functions = Register()
npu_custom_functions([
    npu_apply_adam_w, npu_confusion_transpose, npu_fast_gelu, npu_layer_norm_eval, npu_linear, npu_fusion_attention,
    npu_rms_norm, npu_rotary_mul, npu_scaled_masked_softmax, npu_swiglu, gpu_fusion_attention, npu_apply_adam,
    npu_group_norm_silu, npu_mish, npu_moe_gating_top_k_softmax, npu_sort_v2
])

# register for npu custom backward bench functions
npu_custom_grad_functions = Register()
npu_custom_grad_functions([
    npu_confusion_transpose_backward, npu_fast_gelu_backward, npu_linear_backward, matmul_backward,
    npu_fusion_attention_grad, npu_rms_norm_backward, npu_rotary_mul_backward, npu_scaled_masked_softmax_backward,
    npu_swiglu_backward
])



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.pytorch.common.log import logger
from msprobe.pytorch.dump.module_dump.module_processor import ModuleProcessor
from msprobe.pytorch.dump.api_dump.api_register import get_api_register


class ModuleDumper:
    def __init__(self, service):
        self.service = service
        self.api_register = get_api_register()

    def start_module_dump(self, module, dump_name):
        if hasattr(module, 'msprobe_hook') and not hasattr(module, 'msprobe_module_dump'):
            logger.info_on_rank_0("The init dump is enabled, and the module dump function will not be available.")
            return

        ModuleProcessor.enable_module_dump = True
        self.api_register.restore_all_api()
        if not hasattr(module, 'msprobe_module_dump'):
            self.service.module_processor.register_module_hook(module, self.service.build_hook,
                                                               recursive=False, module_names=[dump_name])
            setattr(module, 'msprobe_module_dump', True)

    def stop_module_dump(self):
        ModuleProcessor.enable_module_dump = False
        self.api_register.register_all_api()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import defaultdict

from msprobe.core.common.file_utils import create_directory, save_json
from msprobe.core.common.const import Const
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger


support_nested_data_type = (list, tuple, dict)


class SingleSave:
    _instance = None

    def __new__(cls, dump_path, fmk=Const.PT_FRAMEWORK):
        if cls._instance is None:
            cls._instance = super(SingleSave, cls).__new__(cls)
            FmkAdp.set_fmk(fmk)
            create_directory(dump_path)

            cls._instance.dump_path = dump_path
            cls._instance.rank = FmkAdp.get_rank_id()
            cls._instance.step_count = 0
            cls._instance.tag_count = defaultdict(int)
        return cls._instance

    @staticmethod
    def _analyze_tensor_data(data, data_name=None, save_dir=None):
        '''
        data: Tensor
        return:
            result_data: with keys {"max", "min", "mean", "norm", "shape"}
        '''
        result_data = {}
        result_data["max"] = FmkAdp.tensor_max(data)
        result_data["min"] = FmkAdp.tensor_min(data)
        result_data["mean"] = FmkAdp.tensor_mean(data)
        result_data["norm"] = FmkAdp.tensor_norm(data)
        result_data["shape"] = list(data.shape)
        if save_dir is not None and data_name is not None:
            real_save_path = os.path.join(save_dir, data_name + ".npy")
            FmkAdp.save_tensor(data, real_save_path)
        return result_data

    @classmethod
    def save_config(cls, data):
        dump_file = os.path.join(cls._instance.dump_path, 'configurations.json')
        save_json(dump_file, data, indent=4)

    @classmethod
    def save_ex(cls, data, micro_batch=None):
        '''
        data: dict{str: Union[Tensor, tuple, list]}

        return: void
        '''

        instance = cls._instance

        if not isinstance(data, dict):
            logger.warning("SingleSave data type not valid, "
                             "should be dict. "
                             "Skip current save process.")
            return
        for key, value in data.items():
            if not isinstance(key, str):
                logger.warning("key should be string when save data")
                continue
            if not isinstance(value, support_nested_data_type) and not FmkAdp.is_tensor(value):
                logger.warning(f"value should be {support_nested_data_type} or Tensor when save data")
                continue
            real_dump_dir = os.path.join(
                instance.dump_path, 
                "data", 
                key, 
                f"step{instance.step_count}", 
                f"rank{instance.rank}")
            if micro_batch is not None:
                real_dump_dir = os.path.join(real_dump_dir, f"micro_step{micro_batch}")
            create_directory(real_dump_dir)

            if FmkAdp.is_tensor(value):
                result = cls._analyze_tensor_data(value, key, real_dump_dir)
            elif isinstance(value, (tuple, list)):
                result = cls._analyze_list_tuple_data(value, key, real_dump_dir)
            elif isinstance(value, dict):
                result = cls._analyze_dict_data(value, key, real_dump_dir)

            result_json = {"data": result}
            json_path = os.path.join(real_dump_dir, key + ".json")
            save_json(json_path, result_json, indent=4)


    @classmethod
    def step(cls):
        instance = cls._instance
        instance.tag_count = defaultdict(int)
        instance.step_count += 1

    @classmethod
    def save(cls, data):
        instance = cls._instance
        if not isinstance(data, dict):
            logger.warning("SingleSave data type not valid, "
                             "should be dict. "
                             "Skip current save process.")
            return
        for key, value in data.items():
            cls.save_ex({key: value}, micro_batch=instance.tag_count[key])
            instance.tag_count[key] += 1

    @classmethod
    def _analyze_list_tuple_data(cls, data, data_name=None, save_dir=None):
        lst = []
        for index, element in enumerate(data):
            if not FmkAdp.is_tensor(element):
                raise TypeError(f"SingleSave: Unsupported type: {type(element)}")
            element_name = data_name + "." + str(index)
            lst.append(cls._analyze_tensor_data(element, element_name, save_dir))
        return lst

    @classmethod
    def _analyze_dict_data(cls, data, data_name=None, save_dir=None):
        result_data = {}
        for key, value in data.items():
            if not FmkAdp.is_tensor(value):
                raise TypeError(f"SingleSave: Unsupported type: {type(value)}")
            key_name = data_name + "." + str(key)
            result_data[key] = cls._analyze_tensor_data(value, key_name, save_dir)
        return result_data
